
#include <numeric>

#include "evaluate.h"
#include "position.h"

std::vector<int16_t> v_pp;
int16_t (*p_pp)[PEndIndex], *p_at;

//Տ̋
constexpr int PPI(const Color c, const PieceType p) {
    return ((int)CvtIfGolds(p) - (CvtIfGolds(p) > HandTypeNum)) * (int)SquareNum + c * BlackPEndIndex;
}
constexpr int PieceIndexTable[ColorNum][PieceTypeNum] = {
    PPI(Black, King), PPI(Black, Rook), PPI(Black, Bishop), PPI(Black, Gold), PPI(Black, Silver), PPI(Black, Knight), PPI(Black, Lance), PPI(Black, Pawn),
    0, PPI(Black, Dragon), PPI(Black, Horse), 0, PPI(Black, ProSilver), PPI(Black, ProKnight), PPI(Black, ProLance), PPI(Black, ProPawn),
    PPI(White, King), PPI(White, Rook), PPI(White, Bishop), PPI(White, Gold), PPI(White, Silver), PPI(White, Knight), PPI(White, Lance), PPI(White, Pawn),
    0, PPI(White, Dragon), PPI(White, Horse), 0, PPI(White, ProSilver), PPI(White, ProKnight), PPI(White, ProLance), PPI(White, ProPawn),
};
forceinline constexpr PIndex PieceIndex(const Color c, const PieceType p, const Square sq) {
    return PieceIndexTable[c][p] + sq;
}
PIndex PieceIndex2(const Color c, const PieceType p, const Square sq) {
    return PieceIndexTable[c][p] + sq;
}

const auto PIndexArrays = [] {
    std::array<std::array<PIndex, PEndIndex>, 4> a;
    for (PIndex pi = 0; pi < PEndIndex; ++pi) {
        a[0][pi] = a[1][pi] = PEndIndex;
    }
    PIndex pi = 0;
    for (Color c = Black; c < ColorNum; ++c) {
        for (PieceType p = King; p < ProGold; ++p) {
            if (p == HandTypeNum) continue;
            for (Square sq = SQ_9A; sq < SquareNum; ++sq) {
                assert(pi == PieceIndex(c, p, sq));
                a[0][pi] = PieceIndex(OppositeColor(c), p, InverseSquare(sq));
                a[1][pi] = PieceIndex(c, p, FlipSquare(sq));
                a[2][pi] = sq;
                a[3][pi] = p;
                ++pi;
            }
        }
        for (PieceType p = Rook; p < HandTypeNum; ++p) {
            const int k = HandIndexSub(p + 1, 0) - HandIndexSub(p, 0);
            for (int i = 0; i < k; ++i) {
                assert(pi == HandIndex(c, p, i));
                a[0][pi] = HandIndex(OppositeColor(c), p, i);
                a[1][pi] = HandIndex(c, p, i);
                a[2][pi] = -1;
                a[3][pi] = p;
                ++pi;
            }
        }
    }
    assert(pi == PEndIndex);
    for (PIndex pi = 0; pi < PEndIndex; ++pi) {
        assert(0 <= a[0][pi] && a[0][pi] < PEndIndex);
        assert(0 <= a[1][pi] && a[1][pi] < PEndIndex);
        assert(a[0][a[0][pi]] == pi);
        assert(a[1][a[1][pi]] == pi);
    }
    return a;
}();
//PIndex̎Ԃƈʒu𔽓] ͎Ԃ𔽓]邾
PIndex InverseIndex(const PIndex pi) {
    return PIndexArrays[0][pi];
}
PIndex FlipIndex(const PIndex pi) {
    return PIndexArrays[1][pi];
}
Square IndexToSquare(const PIndex pi) {
    return static_cast<Square>(PIndexArrays[2][pi]);
}
PieceType IndexToPiece(const PIndex pi) {
    return static_cast<PieceType>(PIndexArrays[3][pi]);
}

bool fv_is_ok() {
    int m = 0;
    //PP
    for (PIndex p1 = 0; p1 < PEndIndex; ++p1) {
        if (p_pp[p1][p1] != 0) m |= 1;
        for (PIndex p2 = 0; p2 < PEndIndex; ++p2) {
            int16_t v1 = p_pp[p1][p2];
            int16_t v2 = p_pp[p2][p1];
            int16_t v3 = p_pp[FlipIndex(p1)][FlipIndex(p2)];
            int16_t v4 = -p_pp[InverseIndex(p1)][InverseIndex(p2)];
            if (v1 != v2) m |= 2;
            if (v1 != v3) m |= 4;
            if (v1 != v4) m |= 8;
            //if (v1 != v4) {
            //    std::cout << v1 << " " << v4 << " " << IndexToSquare(p1) << " " << IndexToPiece(p1) << " " << IndexToSquare(p2) << " " << IndexToPiece(p2) << std::endl;
            //}
        }
    }
    if (m != 0) std::cout << "error fv_is_ok:" << m << std::endl;
    return m == 0;
}

bool list_is_ok(const Position& pos) {
    for (int i = 0; i < ListSize; ++i) {
        const PIndex pi = pos.list_[i];
        const int pih = pi % BlackPEndIndex - (BlackPEndIndex - HandSize);
        if (pih >= 0 && pos.hand_to_list_index_[pi >= BlackPEndIndex][pih] != i) return false;
    }
    for (Color c = Black; c < ColorNum; ++c) {
        bool ret = any_of_bb(pos.byColorBB_[c], [&](const Square sq) {
            PieceType p = pos.piece_[sq];
            return pos.list_[pos.square_to_list_index_[sq]] != PieceIndex(c, p, sq);
        });
        if (ret) return false;
    }
    return true;
}

#define P (o.value)
//XgXVKPP̍vZ Cmove̎ CColorNumȂ獷vZ肩猩XRAԂ posdoMovê
template <Color C>
Score evaluate(Position& pos) {
    const Move move = pos.previous_move_;
    if (C < ColorNum && move.isNone()) {
        //Null moveȂÓI]l͕ςȂ
        assert(abs(pos.forward(-1).static_eval_) < ScoreMateInMaxPly);
        return -pos.forward(-1).static_eval_;
    }

#if 1
    //̗L\rbg{[h
    auto kiki = [&](const Color c) {
        Bitboard b = ZeroBB;
        for_each_piece<true>([&](auto o) {
            Bitboard fromBB = pos.byTypeBB_[P] & pos.byColorBB_[c];
            for_each_bb(fromBB, [&](const Square from) {
                b |= attacks_bb(c, from, P, pos.occupied_);
            });
        });
        return b;
    };
    const Bitboard bb[ColorNum * 2] = { kiki(Black), pos.byColorBB_[Black], kiki(White), pos.byColorBB_[White] };
    constexpr int d[AtNum] = { -FileNum - 1, -FileNum, -FileNum + 1, -1, 0, 1, FileNum - 1, FileNum, FileNum + 1 };
    int sc = 0;
    for (Color c = Black; c < ColorNum; ++c) {
        int king = pos.king_[c];
        int a[AtNum] = {}, sum = 0;
        const int sgn = ColorToSign(c);
        for (int j = 0; j < _countof(bb); ++j) {
            const int j2 = (c == Black) ? j : (j ^ 2);
            const int u = 1 << j2;
            const Bitboard b = bb[j2];
            for (int i = 0; i < _countof(d); ++i) {
                int t = king + d[i] * sgn;
                if (t < 0 || t >= SquareNum || king % FileNum - t % FileNum > 2) a[i] = 15;
                else if (b.isSet((Square)t)) a[i] |= u;
            }
        }
        for (int j = 0; j < AtNum; ++j) {
            const int r = (j * 16 + a[j]) * AtEndIndex;
            for (int i = 0; i < j; ++i) {
                sum += p_at[r + i * 16 + a[i]];
            }
        }
        sc += sum * sgn;
    }
    pos.static_eval_raw_[1] = sc;
#endif

    PIndex *const list = pos.list_;

    //݂̕]l𒼐ڌvZ ƃfobOp
    auto direct_evaluate = [&]() {
        int sum[3] = {};
        //Xg
        int list_index = 0;
        for (Color c = Black; c < ColorNum; ++c) {
            for_each_bb(pos.byColorBB_[c], [&](const Square sq) {
                PieceType p = pos.piece_[sq];
                list[list_index] = PieceIndex(c, p, sq);
                pos.square_to_list_index_[sq] = list_index++;
                if (p != King) sum[2] += PieceScore[p] * ColorToSign(c);
            });
            for (PieceType p = Rook; p < HandTypeNum; ++p) {
                for (int i = 0; i < pos.hand_[c][p]; ++i) {
                    list[list_index] = HandIndex(c, p, i);
                    pos.hand_to_list_index_[c][HandIndexSub(p, i)] = list_index++;
                    sum[2] += PieceScore[p] * ColorToSign(c);
                }
            }
        }
        assert(list_index == ListSize);
        //XgPPvZ
        for (int j = 0; j < ListSize; ++j) {
            auto p = p_pp[list[j]];
            for (int i = 0; i < j; ++i) {//P1P1̓[Ȃ̂œȂ
                sum[0] += p[list[i]];
            }
        }
        pos.static_eval_raw_[0] = sum[0];
        pos.static_eval_raw_[2] = sum[2];
        return static_cast<Score>(pos.static_eval_raw_[0] / FVScale) + pos.static_eval_raw_[1] / ATScale + pos.static_eval_raw_[2];
    };

    if (C == ColorNum) {
        return direct_evaluate();
    }

    //pi0pi1
    auto move_one_piece = [&](const PIndex pi0, const PIndex pi1, int list_index) {
        int16_t *pp0 = p_pp[pi0], *pp1 = p_pp[pi1];

        assert(list[list_index] == pi0);
        list[list_index] = pi1;//ł͎gȂ̂łXVĂ
        int sum = 0;

        for (int i = 0; i < list_index; ++i) {
            sum += pp1[list[i]] - pp0[list[i]];
        }
        for (int i = list_index + 1; i < ListSize; ++i) {
            sum += pp1[list[i]] - pp0[list[i]];
        }
        sum += pp1[pi1] - pp0[pi0];//ɋƂ̉lƂꂽ

        pos.static_eval_raw_[0] += sum;
    };

    const Square from = move.from();
    const Square to = move.to();

    if (from < SquareNum) {
        //ړ
        const PieceType cap = move.cap();

        if (cap) {
            const PieceType hand = cap % HandTypeNum;
            const PIndex pih = HandIndexSub(hand, pos.hand_[C][hand] - 1);

            int list_index = pos.hand_to_list_index_[C][pih] = pos.square_to_list_index_[to];
            assert(list[list_index] == PieceIndex(OppositeColor(C), cap, to));
            move_one_piece(list[list_index], HandIndex(C, pih), list_index);
        }
        int list_index = pos.square_to_list_index_[to] = pos.square_to_list_index_[from];
        assert(list[list_index] == PieceIndex(C, move.piece(), from));
        move_one_piece(list[list_index], PieceIndex(C, move.pieceTo(), to), list_index);

    } else {
        //ł
        const PieceType piece = static_cast<PieceType>(from - SquareNum);

        int list_index = pos.square_to_list_index_[to] = pos.hand_to_list_index_[C][HandIndexSub(piece, pos.hand_[C][piece])];
        assert(list[list_index] == HandIndex(C, piece, pos.hand_[C][piece]));
        move_one_piece(list[list_index], PieceIndex(C, piece, to), list_index);
    }

    Score score = (static_cast<Score>(pos.static_eval_raw_[0] / FVScale) + pos.static_eval_raw_[1] / ATScale + pos.static_eval_raw_[2]) * ColorToSign(OppositeColor(C));
    assert(list_is_ok(pos));
    assert(score == direct_evaluate() * ColorToSign(OppositeColor(C)));//ĂԂтɋXg𐳂Ⴄ悭Ȃ
    return score;
}
template Score evaluate<Black>(Position& pos);
template Score evaluate<White>(Position& pos);
template Score evaluate<ColorNum>(Position& pos);
