
#include <iomanip>
#include <fstream>
#include <vector>
#include <string>
#include <algorithm>
#include <random>
#include <thread>
#include <mutex>
#include <atomic>

#include "learn.h"

#ifdef LEARN
#include "bitboard.h"
#include "search.h"
#include "thread.h"
#include "movegen.h"
#include "tt.h"
#include "util.h"

struct Features {
    Score static_eval_;
    Move move_;
    char is_pv_;
    Square king_[ColorNum];
    PIndex list_[ListSize];
    int a[ColorNum][AtNum];
};

std::mutex mutex;

int16_t *b_pp/*, *b_ppr*/;//ŃyieB̊lƂ]xNg
int16_t *t_pp, *t_ppr/*, *t_at*/;//΂Ƒ΂Ɨ̕]xNg
double *s_pp, *s_at;//ςƂ邽߂̍v
int16_t *w_pp;//ۑp
float *g_pp, *g_ppr, *g_at, *g2_pp, *g2_ppr, *g2_at;//2֌W̌zA2΂̌zAz̓a
float g_material[PieceTypeNum];//̌z
int8_t *f_pp, *f_at;//ŎgtO

inline int pp_index(const PIndex pi0, const PIndex pi1) {
    return pi0 + pi1 * PEndIndex;
}
inline int kp_index(Square king_black, PIndex pi) {
    return (int)king_black * PEndIndex + pi;
}

//ő
constexpr int MaxLearn = MaxMove;

//constexpr Score Window = static_cast<Score>(256);
double J;

forceinline float sigmoid(const float x) {
    return 1.f / (1.f + exp(-0.025f * x));
}
//Ƃ̃XP[VOChƍĂȂȁHJ̌vZɉe
forceinline float hogemoid(float x) {
    x *= 0.01f;
    return x - log(x + 1.f);
}
forceinline float dhogemoid(float x) {
    return x / (x + 100.f);
}

#define P (o.value)
//u\H薖[ǖʂ̓𓾂
void leaf_features(Features& f, const Position& pos_, const RootMove& rm) {
    Position pos = pos_;
    Color c = pos.root_turn_;
    int ply;
    pos.this_thread_->root_moves_[0].score_for_learn_;
    
    for (ply = 0; ply < 32; ++ply) {
        Move move = rm.pv_[ply];
        if (move.isNone()) break;

        pos.doMoveInPlace(c, move);
        c = OppositeColor(c);
    }

    if (ply == 0/* || ply > 32*/) {
        std::cout << "ply:" << ply << std::endl;
        std::cout << "leaf pos:" << pos.toUSI() << std::endl;
    }

    evaluate<ColorNum>(pos);

    auto kiki = [&](const Color c) {
        Bitboard b = ZeroBB;
        for_each_piece<true>([&](auto o) {
            Bitboard fromBB = pos.byTypeBB_[P] & pos.byColorBB_[c];
            for_each_bb(fromBB, [&](const Square from) {
                b |= attacks_bb(c, from, P, pos.occupied_);
            });
        });
        return b;
    };
    const Bitboard bb[ColorNum * 2] = { kiki(Black), pos.byColorBB_[Black], kiki(White), pos.byColorBB_[White] };
    constexpr int d[AtNum] = { -FileNum - 1, -FileNum, -FileNum + 1, -1, 0, 1, FileNum - 1, FileNum, FileNum + 1 };
    for (Color c = Black; c < ColorNum; ++c) {
        int king = pos.king_[c];
        int a[AtNum] = {}, sum = 0;
        const int sgn = ColorToSign(c);
        for (int j = 0; j < _countof(bb); ++j) {
            const int j2 = (c == Black) ? j : (j ^ 2);
            const int u = 1 << j2;
            const Bitboard b = bb[j2];
            for (int i = 0; i < _countof(d); ++i) {
                int t = king + d[i] * sgn;
                if (t < 0 || t >= SquareNum || king % FileNum - t % FileNum > 2) a[i] = 15;
                else if (b.isSet((Square)t)) a[i] |= u;
            }
        }
        std::memcpy(f.a[c], a, sizeof a);
    }

    f.move_ = rm.pv_[0];
    f.static_eval_ = rm.score_for_learn_;
    f.is_pv_ = (rm.score_ != -ScoreInfinite);
    //f.static_eval_ = -evaluate<ColorNum>(pos) * ColorToSign(pos.root_turn_);//ŋXg
    std::memcpy(f.king_, pos.king_, sizeof f.king_);
    std::memcpy(f.list_, pos.list_, sizeof f.list_);
    insertion_sort(f.list_, f.list_ + _countof(f.list_));
}

template <Color C>
void learn_pos(Position& pos, const Move *const learn_move) {
    ++nodes[5];

    new_search();
    id_loop<C>(pos, 5);
    auto& rms = pos.this_thread_->root_moves_;
    if (rms.size() == 0) return;
    std::stable_sort(rms.begin(), rms.end(), [](const RootMove& rm1, const RootMove& rm2) {
        return rm1.score_for_learn_ > rm2.score_for_learn_;
    });
    //std::cout << std::endl;
    //for (int i = 0; i < rms.size(); ++i) {
    //    std::cout << "rm:" << rms[i].score_ << " " << rms[i].ply_ << std::endl;
    //}
    if (abs(rms[0].score_for_learn_) >= ScoreMateInMaxPly) return;

    Features feat[(MaxLearn + 1)];
    int feat_index = 0;

    ++nodes[4];
    for (auto& rm : rms) {
        if (abs(rm.score_for_learn_) >= ScoreMateInMaxPly) continue;
        leaf_features(feat[feat_index++], pos, rm);
    }

    std::lock_guard<std::mutex> lock(mutex);
    //t̖[̓deltaZA̖̎[̓deltaZ
    auto ff = [](Features& f, float delta) {
        PIndex *const list_0 = f.list_;

        for (int j = 0; j < ListSize; ++j) {
            //
            const PieceType p = IndexToPiece(list_0[j]);
            g_material[p] += delta * ((list_0[j] < BlackPEndIndex) ? 1 : -1);
            //PP
            //float *const kpp_black = g_kpp + kpp_index(king_black, list_0[j], 0);
            //float *const kpp_white = g_kpp + kpp_index(king_white, list_1[j], 0);
            float *const pp = g_pp + pp_index(0, list_0[j]);
            for (int i = 0; i < j; ++i) {//P1P1̓[Ȃ̂œȂ
                pp[list_0[i]] += delta;
            }
        }
        for (Color c = Black; c < ColorNum; ++c) {
            const int *const a = f.a[c];
            const int sgn = ColorToSign(c);
            for (int j = 0; j < AtNum; ++j) {
                const int r = (j * 16 + a[j]) * AtEndIndex;
                for (int i = 0; i < j; ++i) {
                    g_at[r + i * 16 + a[i]] += delta * sgn;
                }
            }
        }
    };
    float sum_delta = 0;
    int ii = -1;
    for (int i = 0; i < feat_index; ++i) {
        //if ((learn_move[0].noScore() == feat[i].move_.noScore()) != (learn_move[0] == feat[i].move_)) std::cout << "hoge  " << feat[i].move_.toUSI() << std::endl;
        if (learn_move[0].noScore() == feat[i].move_.noScore()) { ii = i; break; }
    }
    if (ii == -1) return;
    for (int i = 0; i < feat_index; ++i) {
        if (learn_move[0] == feat[i].move_) continue;
        Score score = feat[i].static_eval_;
        if (abs(score) >= ScoreMateInMaxPly || !feat[i].is_pv_) continue;
        float score_diff = static_cast<float>(feat[i].static_eval_ - feat[ii].static_eval_);
        if (score_diff > 0) ++nodes[1];//tƈ tƓ]le
        if (score_diff >= 0) ++nodes[2];

        //if (alpha < score && score < beta) {
            float sg = sigmoid(score_diff);
            J += sg;
            ++nodes[3];

            ////󂢒TƃXRA߂Ȃ悤
            //float score_diff1 = static_cast<float>(feat[i].static_eval_ - feat[i + 1].static_eval_);
            //J += hogemoid(abs(score_diff1));
            //const float delta1 = ColorToSign(C) * Sign(score_diff1) * dhogemoid(abs(score_diff1)) * 0.1f;
            //ff(feat[i], -delta1);

            const float delta = ColorToSign(C) * sg * (1 - sg);
            ff(feat[i], -delta);
            sum_delta += delta;

        //} else {
            //J += (score < beta) ? sigmoid(static_cast<float>(-Window)) : sigmoid(static_cast<float>(Window));
        //}
    }
    if (sum_delta != 0) ff(feat[ii], sum_delta);
}

//2
template <class F>
void for_bound(const Square sq0, const Square sq1, F f) {
    int xa, ya, xb, yb;
    xa = xb = sq0 % FileNum;
    ya = yb = sq0 / FileNum;

    const int x1 = sq1 % FileNum, y1 = sq1 / FileNum;
    xa = std::min(xa, x1);
    ya = std::min(ya, y1);
    xb = std::max(xb, x1);
    yb = std::max(yb, y1);

    for (int y = -ya; y < RankNum - yb; ++y) {
        for (int x = -xa; x < FileNum - xb; ++x) {
            f(x + y * (int)FileNum);
        }
    }
}
//Modeɉēlނ̊evffs
//PPn 0:PP 1:PPΏ̐ 2:PP
template <int Mode, class F>
void update(const PIndex pi0, const PIndex pi1, F f) {
    auto g = [&](const PIndex pi0, const PIndex pi1) {
        f(pp_index(pi0, pi1), 1);
        f(pp_index(pi1, pi0), 1);
        f(pp_index(FlipIndex(pi0), FlipIndex(pi1)), 1);
        f(pp_index(FlipIndex(pi1), FlipIndex(pi0)), 1);
        f(pp_index(InverseIndex(pi0), InverseIndex(pi1)), -1);
        f(pp_index(InverseIndex(pi1), InverseIndex(pi0)), -1);
        f(pp_index(InverseIndex(FlipIndex(pi0)), InverseIndex(FlipIndex(pi1))), -1);
        f(pp_index(InverseIndex(FlipIndex(pi1)), InverseIndex(FlipIndex(pi0))), -1);
    };
    if (Mode == 0) {
        //PP
        f(pp_index(pi0, pi1), 1);

    } else if (Mode == 1) {
        //PPEΏ2ւ] PP͊ƂȂʂȂ̂ŉ]Kv
        g(pi0, pi1);

    } else if (Mode == 2) {
        //PP
        for_bound(IndexToSquare(pi0), IndexToSquare(pi1), [&](const int d) {
            f(pp_index(pi0 + d, pi1 + d), 1);
        });
    }
}


template <class F>
void for_pp(F f) {
    for (PIndex pi0 = 0; pi0 < PEndIndex; ++pi0) {
        for (PIndex pi1 = 0; pi1 < PEndIndex; ++pi1) {
            f(pi0, pi1);
        }
    }
}
template <class F>
void for_kp(F f) {
    for (Square king_black = SQ_9A; king_black < SquareNum; ++king_black) {
        for (PIndex pi = 0; pi < PEndIndex; ++pi) {
            f(king_black, pi);
        }
    }
}


//zƃyieB̑ΏۂƃyieB̊lۂ̕ϓԂ
inline int16_t get_h(const float g0, const float v, float& h2, const float b, const int m) {
    const float d = v - b;
    int d_sgn = (d == 0) ? 0 : Sign(d);//yieB̌
    //float g = g0 - d_sgn * log((float)(abs(d * 2) + 1)) * 0.25f;//yieBKp̌z
    float g = g0 - d * 0.00625f;//yieBKp̌z
    if (g == 0) return 0;
    int16_t h = Sign(g);
    h2 += 1;// h * h;
    h = (int16_t)(h * m / sqrt(h2));
    if (h == 0) h = Sign(g);
    if (g * g0 < 0 && ((d > 0 && d + h < 0) || (d < 0 && d + h > 0))) h = static_cast<int16_t>(-d);//yieBŕςȂ悤
    //std::cout << h << " " << g << std::endl;
    return h;
}

#define FIRST

void learn(const std::vector<PositionLearn>&& v_posl) {
    std::thread thread[MaxThreads];
    const int ThreadNum = std::thread::hardware_concurrency();
    std::cout << "threads:" << ThreadNum << " pos_num:" << v_posl.size() << std::endl;
    thread_num = 1;
    set_low_priority();

    int dist[SquareNum][SquareNum];
    for (int sq0 = 0; sq0 < SquareNum; ++sq0) {
        for (int sq1 = 0; sq1 < SquareNum; ++sq1) {
            const int x0 = sq0 % FileNum, y0 = sq0 / FileNum;
            const int x1 = sq1 % FileNum, y1 = sq1 / FileNum;
            dist[sq0][sq1] = std::max(abs(x0 - x1), abs(y0 - y1));
        }
    }

    auto alloc = [&](auto& p, size_t num) {
        p = static_cast<std::remove_reference<decltype(p)>::type>(std::malloc(num * sizeof *p));
    };
    alloc(s_at, ATSize);
    alloc(g_at, ATSize);
    alloc(g2_at, ATSize);
    alloc(b_pp, PPSize);
    alloc(s_pp, PPSize);
    alloc(w_pp, PPSize + ATSize);
    alloc(t_pp, PPSize);
    alloc(t_ppr, PPSize);
    alloc(g_pp, PPSize);
    alloc(g_ppr, PPSize);
    alloc(g2_pp, PPSize);
    alloc(g2_ppr, PPSize);
    alloc(f_pp, PPSize);
    alloc(f_at, ATSize);

#ifdef FIRST
    std::memset(b_pp, 0, PPSize * sizeof *b_pp);
    for (Square kb = SQ_9A; kb < SquareNum; ++kb) {
        for (Square kw = SQ_9A; kw < SquareNum; ++kw) {
            constexpr int16_t v_king[RankNum] = { 650, 600, 500, 300, 150, 50, 10, 0, 0 };//ʕ␳
            if (kb == kw) continue;
            int index = pp_index(PieceIndex2(Black, King, kb), PieceIndex2(White, King, kw));
            int index2 = pp_index(PieceIndex2(White, King, kw), PieceIndex2(Black, King, kb));
            b_pp[index] = b_pp[index2] = (v_king[kb / FileNum] - v_king[RankNum - 1 - kw / FileNum]) * FVScale / 2;
        }
    }
    std::memset(p_pp, 0, PPSize * sizeof **p_pp);
    std::memset(p_at, 0, ATSize * sizeof *p_at);
    std::memcpy(t_pp, b_pp, PPSize * sizeof *t_pp);
#else
    //std::memcpy(b_pp, p_pp, PPSize * sizeof **p_pp);
    std::memset(b_pp, 0, PPSize * sizeof *b_pp);
    std::memcpy(t_pp, p_pp, PPSize * sizeof *t_pp);
#endif
    std::memset(s_pp, 0, PPSize * sizeof *s_pp);
    std::memset(s_at, 0, ATSize * sizeof *s_at);
    //std::memcpy(t_pp, b_pp, PPSize * sizeof *t_pp);
    std::memset(t_ppr, 0, PPSize * sizeof *t_ppr);
    //std::memset(t_at, 0, ATSize * sizeof *t_at);
    std::memset(g2_pp, 0, PPSize * sizeof *g2_pp);
    std::memset(g2_ppr, 0, PPSize * sizeof *g2_ppr);
    std::memset(g2_at, 0, ATSize * sizeof *g2_at);

    threads = static_cast<Thread *>(std::malloc(ThreadNum * sizeof Thread));


    const int PosNum = std::min((int)v_posl.size(), 6000);
    std::mt19937_64 mt(1245);
    int *pp;
    alloc(pp, v_posl.size());
    for (int i = 0; i < v_posl.size(); ++i) pp[i] = i;

    for (int l = 1;; ++l) {
        time_start = std::chrono::high_resolution_clock::now();
        auto time = [&]() {
            const auto duration = std::chrono::high_resolution_clock::now() - time_start;
            const auto msec = std::chrono::duration_cast<std::chrono::milliseconds>(duration).count();
            return msec;
        };

        std::cout << "l:" << l << std::endl;
        std::memset(g_pp, 0, PPSize * sizeof *g_pp);
        std::memset(g_ppr, 0, PPSize * sizeof *g_ppr);
        std::memset(g_at, 0, ATSize * sizeof *g_at);
        std::memset(g_material, 0, sizeof g_material);
        std::memset(nodes, 0, sizeof nodes);
        J = 0;
        std::shuffle(pp, pp + v_posl.size(), mt);

        //]xNǧz߂
        {
            std::atomic_int i0 = 0;
            for (int n = 0; n < ThreadNum; ++n) {
                thread[n] = std::thread([&](const int n1) {
                    Thread *const thisThread = &threads[n1];
                    Position& pos = thisThread->pos_[800];
                    std::memset(thisThread, 0, sizeof *thisThread);

                    thisThread->idx_ = 0;
                    thisThread->ppos_ = &pos;


                    //threads[n1].thread_ = nullptr;
                    std::memset(&pos - 8, 0, 8 * sizeof Position);
                    pos.this_thread_ = thisThread;

                    for (int i; (i = i0++) < PosNum;) {
                        if (i % 1024 == 0) std::cout << i << " ";

                        pos.clear();
                        pos.next().clear();
                        pos.next().next().clear();
                        pos.this_thread_ = thisThread;

                        const PositionLearn& posl = v_posl[pp[i]];
                        std::memcpy(pos.byColorBB_, posl.byColorBB_, sizeof posl.byColorBB_);
                        std::memcpy(pos.hand64_, posl.hand64_, sizeof posl.hand64_);
                        std::memcpy(pos.piece_, posl.piece_, sizeof posl.piece_);
                        pos.root_turn_ = posl.root_turn_;
                        pos.normalize();

                        (pos.root_turn_ == Black) ? learn_pos<Black>(pos, posl.learn_move_) : learn_pos<White>(pos, posl.learn_move_);
                    }
                }, n);
            }
            for (int n = 0; n < ThreadNum; ++n) {
                thread[n].join();
            }
            std::cout << PosNum << "\n";
        }

        std::cout << time() << " search end." << std::endl;

        //Ώ̐gĒlW߂
        std::memset(f_pp, 0, PPSize * sizeof *f_pp);
        for_pp([&](const PIndex pi0, const PIndex pi1) {
            const int index = pp_index(pi0, pi1);
            if (f_pp[index]) return;

            float sum = 0;
            update<1>(pi0, pi1, [&](const int index, const int sgn) {
                if (f_pp[index]) return;
                sum += g_pp[index] * sgn;
                f_pp[index] = 1;
            });
            if (pi0 == InverseIndex(pi1) || FlipIndex(pi0) == InverseIndex(pi1)) sum = 0;
            update<1>(pi0, pi1, [&](const int index, const int sgn) {
                if (f_pp[index] != 1) return;
                g_pp[index] = sum * sgn;
                f_pp[index] = 2;
            });
        });
        std::memset(f_at, 0, ATSize * sizeof *f_at);
        constexpr int rs[AtNum] = { 2,1,0,5,4,3,8,7,6 };
        for (int i = 0; i < ATSize; i++) {
            if (f_at[i]) continue;
            float sum = 0;
            int i0 = i % AtEndIndex, i1 = i / AtEndIndex;
            int s0 = i0 / 16, s1 = i1 / 16;
            int t0 = i0 % 16, t1 = i1 % 16;
            auto f0 = [&](const int s0, const int t0, const int s1, const int t1) {
                int i = s0 * 16 + t0 + (s1 * 16 + t1) * AtEndIndex;
                if (f_at[i]) return;
                sum += g_at[i];
                f_at[i] = 1;
            };
            f0(s0, t0, s1, t1);
            f0(rs[s0], t0, rs[s1], t1);
            f0(s1, t1, s0, t0);
            f0(rs[s1], t1, rs[s0], t0);
            auto f1 = [&](const int s0, const int t0, const int s1, const int t1) {
                int i = s0 * 16 + t0 + (s1 * 16 + t1) * AtEndIndex;
                if (f_at[i] != 1) return;
                g_at[i] = sum;
                f_at[i] = 2;
            };
            f1(s0, t0, s1, t1);
            f1(rs[s0], t0, rs[s1], t1);
            f1(s1, t1, s0, t0);
            f1(rs[s1], t1, rs[s0], t0);
        }

        //2΂̌zW߂
        std::memset(f_pp, 0, PPSize * sizeof *f_pp);
        for_pp([&](const PIndex pi0, const PIndex pi1) {
            const int index = pp_index(pi0, pi1);
            if (f_pp[index]) return;
            if (isNegative(IndexToSquare(pi0)) || isNegative(IndexToSquare(pi1))) return;

            float sum = 0;
            update<2>(pi0, pi1, [&](const int index, const int sgn) {
                if (f_pp[index]) return;
                sum += g_pp[index] * sgn;
                f_pp[index] = 1;
            });
            if (IndexToPiece(pi0) == IndexToPiece(pi1) && ((pi0 < BlackPEndIndex) != (pi1 < BlackPEndIndex))) sum = 0;
            update<2>(pi0, pi1, [&](const int index, const int sgn) {
                if (f_pp[index] != 1) return;
                g_ppr[index] = sum * sgn;
                f_pp[index] = 2;
            });
        });

        //@Ƌʂ̊֌W
        for (PIndex pi = BlackPEndIndex - HandSize; pi < BlackPEndIndex; ++pi) {
            float sum = 0;
            for (Square k = SQ_9A; k < SquareNum; ++k) {
                sum += g_pp[pp_index(PieceIndex2(Black, King, k), pi)];
            }
            for (Square k = SQ_9A; k < SquareNum; ++k) {
                int index = pp_index(PieceIndex2(Black, King, k), pi), index2 = pp_index(pi, PieceIndex2(Black, King, k));
                int index3 = pp_index(InverseIndex(PieceIndex2(Black, King, k)), InverseIndex(pi)), index4 = pp_index(InverseIndex(pi), InverseIndex(PieceIndex2(Black, King, k)));
                g_ppr[index] = g_ppr[index2] = sum;
                g_ppr[index3] = g_ppr[index4] = -sum;
            }
        }

        //ꂽmPPȂ
        auto f = [&](const PIndex pi0, const PIndex pi1) {
            PieceType p0 = IndexToPiece(pi0), p1 = IndexToPiece(pi1);
            if (p0 == King || p1 == King) return 250;
            Square sq0 = IndexToSquare(pi0), sq1 = IndexToSquare(pi1);
            if (isNegative(sq0) || isNegative(sq1)) return 30;
            Color c0 = (pi0 < BlackPEndIndex) ? Black : White, c1 = (pi1 < BlackPEndIndex) ? Black : White;
            int x0 = sq0 % FileNum, y0 = sq0 / FileNum;
            int x1 = sq1 % FileNum, y1 = sq1 / FileNum;
            if (p0 == Knight || p0 == Lance) y0 -= ColorToSign(c0);
            if (p1 == Knight || p1 == Lance) y1 -= ColorToSign(c1);
            return (abs(x0 - x1) < 3 && abs(y0 - y1) < 3) ? 30 : 0;
        };

        //zɉĕ]xNgXV
        //PP
        //std::memset(f_pp, 0, PPSize * sizeof *f_pp);
        for_pp([&](const PIndex pi0, const PIndex pi1) {
            const int m = f(pi0, pi1);
            if (m == 0) return;
            const int index = pp_index(pi0, pi1);

            t_pp[index] += get_h(g_pp[index], t_pp[index], g2_pp[index], b_pp[index], m);

            //PP
            t_ppr[index] += get_h(g_ppr[index], t_ppr[index], g2_ppr[index], 0, m);
        });
        std::cout << time() << " pp end." << std::endl;

#ifdef FIRST
        //ʂ痣قǉl锛
        if (l == 0) {
            //ςςȂ
            int a_sum[SquareNum][ColorNum][ProGold];
            for (Square kb = SQ_9A; kb < SquareNum; ++kb) {
                const PIndex pi_king = PieceIndex2(Black, King, kb);
                for (Color c = Black; c < ColorNum; ++c) {
                    for (PieceType p = Rook; p < ProGold; ++p) {
                        if (p == HandTypeNum || IsSlider(p)) continue;
                        int sum = 0, n = 0;
                        for (Square sq = SQ_9A; sq < SquareNum; ++sq) {
                            const PIndex pi = PieceIndex2(c, p, sq);
                            sum += t_ppr[kp_index(kb, pi)];
                        }
                        a_sum[kb][c][p] = sum;
                    }
                }
            }
            for (int i = 0; i < 11; ++i) {
                std::memset(f_pp, 0, PPSize * sizeof *f_pp);
                for_kp([&](const Square kb, const PIndex pi) {
                    if (isNegative(IndexToSquare(pi))) return;//Ƃ̊֌W
                    const int index = kp_index(kb, pi);

                    //8+8̕ 3ȏŉ艿lႭ 3ȏŉ艿lႭ 1ȏŏ艿lႭ
                    const Square sq0 = kb, sq1 = IndexToSquare(pi);
                    if (sq0 == sq1) return;
                    PieceType p1 = IndexToPiece(pi);
                    if (IsSlider(p1)) return;//Ԃ͂ǂ悤c
                    Color c1 = (pi < BlackPEndIndex) ? Black : White;
                    int x0 = sq0 % FileNum, y0 = sq0 / FileNum;
                    int x1 = sq1 % FileNum, y1 = sq1 / FileNum;
                    if (p1 == Knight || p1 == Lance || p1 == Pawn) {
                        if (p1 == Knight && PromotionZone<2>(c1, sq1)) return;
                        if ((p1 == Lance || p1 == Pawn) && PromotionZone<1>(c1, sq1)) return;
                        y1 -= ColorToSign(c1);
                    }

                    int sum = 0, n = 0;
                    const Bitboard b1 = step_attacks_bb(Black, King, kb);
                    for_each_bb(b1, [&](const Square sq) {
                        ++n;
                        sum += t_ppr[kp_index(sq, pi)];
                    });
                    const Bitboard b2 = step_attacks_bb(Black, King, sq1);
                    for_each_bb(b2, [&](const Square sq) {
                        ++n;
                        sum += t_ppr[kp_index(kb, pi - sq1 + sq)];
                    });
                    sum /= n;

                    int dx = x0 - x1, dy = y0 - y1;//ʂǂɂ邩
                    int dist = std::max(abs(dx), abs(dy));//`FXՋ
                    int sgn = ColorToSign(c1);
                    int v = t_ppr[index] * sgn, v0 = v;//pi̎Ԃ猩l

                    if (abs(dx) > 2) {
                        v = std::min(v, t_ppr[index + Sign(dx)] * sgn);
                        v = std::min(v, t_ppr[index - Sign(dx) * PEndIndex] * sgn);
                    }
                    if (dy < -1 || dy > 2) {
                        v = std::min(v, t_ppr[index + Sign(dy) * (int)FileNum] * sgn);
                        v = std::min(v, t_ppr[index - Sign(dy) * (int)FileNum * PEndIndex] * sgn);
                    }
                    if (v != v0) v -= PieceScore[p1] / 16 * ((int)FileNum - dist);
                    if (dist > 3) {
                        v = std::min(v, sum * sgn);
                    }
                    update<1>(PieceIndex2(Black, King, kb), pi, [&](const int index, const int sgn2) {
                        if (f_pp[index]) return;
                        t_ppr[index] = v * sgn * sgn2;
                        f_pp[index] = 1;
                    });
                });
            }
            std::memset(f_pp, 0, PPSize * sizeof *f_pp);
            for (Square kb = SQ_9A; kb < SquareNum; ++kb) {
                const PIndex pi_king = PieceIndex2(Black, King, kb);
                for (Color c = Black; c < ColorNum; ++c) {
                    for (PieceType p = Rook; p < ProGold; ++p) {
                        if (p == HandTypeNum || IsSlider(p)) continue;
                        int sum = 0;
                        for (Square sq = SQ_9A; sq < SquareNum; ++sq) {
                            const PIndex pi = PieceIndex2(c, p, sq);
                            sum += t_ppr[kp_index(kb, pi)];
                        }
                        const int t = (a_sum[kb][c][p] - sum) / (int)SquareNum;
                        for (Square sq = SQ_9A; sq < SquareNum; ++sq) {
                            const PIndex pi = PieceIndex2(c, p, sq);
                            update<1>(pi_king, pi, [&](const int index, const int sgn) {
                                if (f_pp[index]) return;
                                t_ppr[index] += t * sgn;
                                f_pp[index] = 1;
                            });
                        }
                    }
                }
            }
        }
#endif
        for_pp([&](const PIndex pi0, const PIndex pi1) {
            const int index = pp_index(pi0, pi1);
            p_pp[0][index] = t_pp[index] + t_ppr[index];
            s_pp[index] += p_pp[0][index];
        });
        std::cout << time() << " kp end." << std::endl;

        for (int i = 0; i < ATSize; i++) {
            p_at[i] += get_h(g_at[i], p_at[i], g2_at[i], 0, 150);
            s_at[i] += p_at[i];
        }

        ////
        //g_material[King] = g_material[HandTypeNum] = g_material[ProGold] = std::numeric_limits<float>::max();
        //g_material[Pawn] *= 0.2f;
        //int i_vp[PieceTypeNum];
        //for (int i = 0; i < PieceTypeNum; ++i) {
        //    i_vp[i] = i;
        //}
        //insertion_sort(i_vp, i_vp + PieceTypeNum, [&](const int a1, const int a2) {
        //    return g_material[a1] < g_material[a2];
        //});
        //for (int i = 0; i < PieceTypeNum - 3; ++i) {//0 6 12
        //    PieceScore[i_vp[i]] += (i - 6) * abs(i - 6) / 4;//36 25 16 9 4 1 0 
        //}
        //for (int i = 0; i < PieceTypeNum; ++i) {
        //    std::cout << i_vp[i] << " ";
        //}

        //v̕\
        for (PieceType p = King; p < PieceTypeNum; ++p) {
            std::cout << std::setw(5) << PieceScore[p];
        }
        int64_t s1 = 0, s2 = 0, m[2] = { INT16_MAX, INT16_MIN };
        for (int i = 0; i < PPSize; ++i) {
            int64_t t = p_pp[0][i] - b_pp[i];
            s1 += abs(t);
            s2 += t * t;
            if (t < m[0]) m[0] = t;
            if (t > m[1]) m[1] = t;
        }
        std::cout << "\n";
        for (Color c = Black; c < ColorNum; ++c) {
            int sum = 0;
            Square king_black = square(1, 7);
            for (Square sq = SQ_9A; sq < SquareNum; ++sq) {
                if (sq % FileNum == 0) std::cout << '\n';
                const PIndex pi0 = PieceIndex2(c, Gold, sq);
                int s = p_pp[PieceIndex2(Black, King, king_black)][pi0];
                std::cout << std::setw(6) << s;
                sum += s;
            }
            std::cout << " " << sum << "\n\n";
        }
        printf("J:%f failed:%lld %lld learned:%lld %lld %lld fv_is_ok:%d L1:%f L2:%f min:%lld max:%lld\n",
               J, nodes[1], nodes[2], nodes[4], nodes[3], nodes[5], fv_is_ok(), (double)s1 / PPSize, sqrt((double)s2 / PPSize), m[0], m[1]);
        //std::cout << "J:" << J << " failed:" << nodes[1] << " " << nodes[2] << " learned:" << nodes[4] << " " << nodes[3] << " " << nodes[5]
        //    << " fv_is_ok:" << fv_is_ok() << " L1:" << (double)s1 / FVSize << " L2:" << sqrt((double)s2 / FVSize) << " min:" << m[0] << " max:" << m[1]
        //    << std::endl;

        //]xNgۑ
        std::ofstream ofs{ "material.txt", std::ios::out | std::ios::trunc };
        if (ofs.is_open()) {
            for (PieceType p = King; p < PieceTypeNum; ++p) {
                ofs << PieceScore[p] << std::endl;
            }
        }
        init_piece_score(false);
        double t = 1.0 / l;
        for_pp([&](const PIndex pi0, const PIndex pi1) {
            const int index = pp_index(pi0, pi1);
            w_pp[index] = (int16_t)(s_pp[index] * t);
        });
        for (int i = 0; i < ATSize; i++) {
            w_pp[i + PPSize] = (int16_t)(s_at[i] * t);
        }
        if (l % 10 == 0) {
            for (int i = 0; i < 900; i++) {
                std::string path = "pp_" + std::to_string(100 + i) + ".bin";
                std::fstream fs;
                fs.open(path, std::ios::in | std::ios::binary);
                if (fs.is_open()) { fs.close(); continue; }
                fs.open(path, std::ios::out | std::ios::binary);
                if (!fs.is_open()) continue;
                fs.write(reinterpret_cast<char *>(w_pp), (PPSize + ATSize) * sizeof *w_pp);
                break;
            }
        }

        //std::quick_exit(0);

        //]xNgω̂Œu\NA
        tt.clear();

        std::cout << time() << " end." << std::endl;
    }
}

#else

void learn(const std::vector<PositionLearn>&& vpos) {}

#endif
