
#pragma once

#include <string>

#include "bitboard.h"

enum PromoteMode {
    NonPromote, Promote,
};

//from, to, piece, pro, cap, score
constexpr int L0 = 7, L1 = 7, L2 = 4, L3 = 1, L4 = 4, L5 = 9;
constexpr int S0 = 8, S1 = 0, S2 = 15, S3 = 7, S4 = 19, S5 = 23;
//ztemplateɎw肷ƃCeZXɈorz
//constexpr int L5 = { 7, 7, 4, 1, 4 };
//constexpr int S5 = { 8, 0, 15, 7, 19 };

class TTEntry;

//w
class Move {
    static constexpr int32_t NoneValue = 0;
    static constexpr int32_t MaxValue = INT32_MAX;

    int32_t value_;

    constexpr Move(int32_t v) : value_(v) {}

    friend class TTEntry;
    friend class TranspositionTable;

public:
    Square from() const {
        return static_cast<Square>(ExtBits<S0, L0>(value_));
    }
    Square to() const {
        return static_cast<Square>(ExtBits<S1, L1>(value_));
    }
    PieceType piece() const {
        return static_cast<PieceType>(ExtBits<S2, L2>(value_));
    }
    PieceType promote() const {
        return static_cast<PieceType>(ExtBits<S3, L3>(value_));
    }
    PieceType cap() const {
        return static_cast<PieceType>(ExtBits<S4, L4>(value_));
    }
    PieceType pieceTo() const {
        return add_if(piece(), isPromotion(), HandTypeNum);
    }

    bool isDrop() const {
        return from() >= SquareNum;
    }
    bool isPromotion() const {
        return (value_ & MaskBits(S3, L3)) != 0;
    }
    bool isCapture() const {
        return (value_ & MaskBits(S4, L4)) != 0;
    }
    bool isNone() const {
        return value_ == NoneValue;
    }

    bool operator==(const Move a1) const {
        return value_ == a1.value_;
    }
    bool operator!=(const Move a1) const {
        return value_ != a1.value_;
    }
    bool operator>(const Move a1) const {
        return (value_ >> S5) > (a1.value_ >> S5);
    }
    bool operator<(const Move a1) const {
        return (value_ >> S5) < (a1.value_ >> S5);
    }
    Move noScore() const {
        return ExtBits<0, S5>(value_);
    }
    Move noScoreQuiet() const {
        return ExtBits<0, S4>(value_);
    }
    static bool greaterQuiet(const Move a1, const Move a2) {
        return (a1.value_ >> S4) > (a2.value_ >> S4);
    }
    static bool lessValue(const Move a1, const Move a2) {
        return a1.value_ < a2.value_;
    }
    Move togglePromotion() const {
        return value_ ^ 1 << S3;
    }
    int score() const {
        return value_ >> S4;
    }

    //wUSI`ɕϊ
    std::string toUSI() const;

    //ړ(ł̂Ƃ͋+SquareNum)7bit, ړ7bit, ړuOv̋(ł̂Ƃ0)4bit, 1bit, 4bit
    template <PromoteMode PM, bool SC = false>
    static Move makeMove(Square from, Square to, PieceType piece, PieceType cap) {
        return add_if(from << S0 | to << S1 | piece << S2 | PM << S3 | cap << S4,
                      SC, ((PieceScore[cap] >> 3) - Lva(piece)) << S5);
    }
    template <PromoteMode PM, bool SC = false>
    static Move makeMove(Square from, Square to, PieceType piece) {
        return add_if(from << S0 | to << S1 | piece << S2 | PM << S3,
                      SC, -Lva(piece) << S5);
    }
    template <PromoteMode PM>
    static Move makeMove(Square from, Square to, PieceType piece, int32_t sc) {
        return makeMove<PM, false>(from, to, piece).value_ | sc << S4;//ł͂Ȃ̂13bit̗]T
    }
    static Move makeDrop(PieceType hand, Square to) {
        return (static_cast<Square>(hand) + SquareNum) << S0 | to << S1 | hand << S2;
    }
    static Move makeDrop(PieceType hand, Square to, int32_t sc) {
        return makeDrop(hand, to).value_ | sc << S4;
    }

    static constexpr Move None() {
        return Move{ NoneValue };
    }
    static constexpr Move MaxScore() {
        return Move{ MaxValue };
    }

    Move() {}
};
