
#include "movegen.h"
#include "bitboard.h"
#include "move.h"
#include "position.h"
#include "search.h"
#include "thread.h"

#define P (o.value)
#define PFrom ((P == Golds) ? pos.piece_[from] : P)

template <Color C, bool SC>
Move *generate_captures(Move *moves, const Position& pos) {
    //܂͋ȊO𐬂
    for_each_piece([&](auto o) {
        for_each_bb(pos.byTypeBB_[P] & pos.byColorBB_[C], [&](const Square from) {
            Bitboard attackBB = attacks_bb(C, from, P, pos.occupied_);
            if (P == King) attackBB &= ~pos.checkerAttacksBB_;

            //
            for_each_bb(attackBB & pos.byColorBB_[OppositeColor(C)], [&](const Square to) {
                const PieceType cap = pos.piece_[to];
                if (CanPromote(P) && (PromotionZone(C, from) || PromotionZone(C, to))) {
                    *moves++ = Move::makeMove<Promote, SC>(from, to, PFrom, cap);
                }
                if (P == Knight && PromotionZone<4>(C, from)) return;
                if (P == Lance && PromotionZone<2>(C, to)) return;
                if (P == Pawn && PromotionZone<PromotionRank + 1>(C, from)) return;
                if ((P != Rook && P != Bishop && P != Pawn) || !(PromotionZone(C, from) || PromotionZone(C, to))) {
                    *moves++ = Move::makeMove<NonPromote, SC>(from, to, PFrom, cap);
                }
            });

            //Ȃ肩𐬂
            if (P == Pawn) {
                for_each_bb(attackBB & pos.byTypeBB_[Empty], [&](const Square to) {
                    if (PromotionZone(C, from) || PromotionZone(C, to)) {
                        *moves++ = Move::makeMove<Promote, SC>(from, to, PFrom);
                    }
                });
            }
        });
    });
    return moves;
}
template Move *generate_captures<Black, true>(Move *moves, const Position& pos);
template Move *generate_captures<White, true>(Move *moves, const Position& pos);

template <Color C>
Move *generate_non_captures(Move *moves, const Position& pos) {
    //Ȃ肩𐬂Ȃ
//    const auto history = pos.this_thread_->history_;
    for_each_piece([&](auto o) {
        Bitboard fromBB = pos.byTypeBB_[P] & pos.byColorBB_[C];
        //if (P == Knight) fromBB &= InFrontBB[OppositeColor(C)][RankNum - 4];

        for_each_bb(fromBB, [&](const Square from) {
            Bitboard attackBB = attacks_bb(C, from, P, pos.occupied_) & pos.byTypeBB_[Empty];
            if (P == King) attackBB &= ~pos.checkerAttacksBB_;
            //if (P == Lance) attackBB &= InFrontBB[OppositeColor(C)][RankNum - 2];

            for_each_bb(attackBB, [&](const Square to) {
                if (CanPromote(P) && P != Pawn && (PromotionZone(C, from) || PromotionZone(C, to))) {
                    *moves++ = Move::makeMove<Promote>(from, to, PFrom, history.value(C, false, PFrom, to));
                }
                if (P == Knight && PromotionZone<4>(C, from)) return;
                if (P == Lance && PromotionZone<2>(C, to)) return;
                if (P == Pawn && PromotionZone<PromotionRank + 1>(C, from)) return;
                if ((P != Rook && P != Bishop && P != Pawn) || !(PromotionZone(C, from) || PromotionZone(C, to))) {
                    *moves++ = Move::makeMove<NonPromote>(from, to, PFrom, history.value(C, false, PFrom, to));
                }
            });
        });
    });
    return moves;
}

template <Color C>
Move *generate_drops(Move *moves, const Position& pos) {
    //ł
    //const auto history = pos.this_thread_->history_;
    for_each_piece([&](auto o) {
        if (P == King || P >= HandTypeNum) return;
        if (pos.hand_[C][P] == 0) return;

        Bitboard targetBB = pos.byTypeBB_[Empty];
        if (P == Pawn) targetBB &= pawn_file_bb(C, pos);
        if (P == Pawn || P == Lance) targetBB &= InFrontBB[OppositeColor(C)][RankNum - 1];
        if (P == Knight) targetBB &= InFrontBB[OppositeColor(C)][RankNum - 2];

        for_each_bb(targetBB, [&](const Square to) {
            *moves++ = Move::makeDrop(P, to, history.value(C, true, P, to));
        });
    });
    return moves;
}

template <Color C>
Move *generate_recaptures(Move *moves, const Position& pos) {
    //Lv` previous_move_Ȃꍇ͌Ă΂Ȃ낤
    const Square to = pos.previous_move_.to();
    const PieceType cap = pos.piece_[to];
    Bitboard fromBB = pos.attackersTo<C>(to, pos.occupied_);
    for_each_bb(fromBB, [&](const Square from) {
        const PieceType p = pos.piece_[from];
        if (CanPromote(p) && (PromotionZone(C, to) || PromotionZone(C, from))) {
            *moves++ = Move::makeMove<Promote>(from, to, p, cap);
        } else {
            *moves++ = Move::makeMove<NonPromote>(from, to, p, cap);
        }
    });
    return moves;
}
template Move *generate_recaptures<Black>(Move *moves, const Position& pos);
template Move *generate_recaptures<White>(Move *moves, const Position& pos);

template <Color C>
Move *generate_non_evasions(Move *moves, const Position& pos) {
    moves = generate_captures<C, false>(moves, pos);
    Move *non_captures_first = moves;

    moves = generate_non_captures<C>(moves, pos);
    moves = generate_drops<C>(moves, pos);
    std::for_each(non_captures_first, moves, [](Move& move) {
        move = move.noScoreQuiet();
    });
    return moves;
}

template <Color C>
Move *generate_non_evasions_ex(Move *moves, const Position& pos) {
    //p̕sA2iڂ̍̕s
    for_each_piece([&](auto o) {
        if (P != Rook && P != Bishop && P != Lance && P != Pawn) return;

        Bitboard fromBB = pos.byTypeBB_[P] & pos.byColorBB_[C];
        if (P == Pawn) fromBB &= RankBB[RankInv(C, 2)] ^ RankBB[RankInv(C, 3)];//3,4i

        for_each_bb(fromBB, [&](const Square from) {
            Bitboard attackBB = attacks_bb(C, from, P, pos.occupied_) & ~pos.byColorBB_[C];
            if (P == Lance) attackBB &= RankBB[RankInv(C, 1)];//2i

            for_each_bb(attackBB, [&](const Square to) {
                if (P == Pawn || P == Lance || (PromotionZone(C, from) || PromotionZone(C, to))) {
                    const PieceType cap = pos.piece_[to];
                    *moves++ = Move::makeMove<NonPromote>(from, to, P, cap);
                }
            });
        });
    });
    return moves;
}

template <Color C>
Move *generate_evasions_base(Move *moves, const Position& pos) {
    const PieceType cap = pos.piece_[pos.checker_];

    if (pos.checked_ != 1) {
        //ʂ𓮂 ̂Ƃ͂ꂾ
        Bitboard toBB = attacks_bb(C, pos.king_[C], King, pos.occupied_) & ~pos.byColorBB_[C] & ~pos.checkerAttacksBB_;
        for_each_bb(toBB, [&](const Square to) {
            *moves++ = Move::makeMove<NonPromote, true>(pos.king_[C], to, King, pos.piece_[to]);
        });
        return moves;
    }

    //checker
    for_each_piece([&](auto o) {
        Bitboard fromBB = attacks_bb(OppositeColor(C), pos.checker_, P, pos.occupied_) & pos.byTypeBB_[P] & pos.byColorBB_[C];
        for_each_bb(fromBB, [&](const Square from) {
            if (CanPromote(P) && (PromotionZone(C, pos.checker_) || PromotionZone(C, from))) {
                *moves++ = Move::makeMove<Promote, true>(from, pos.checker_, P, cap);
                if (P == Rook || P == Bishop || P == Pawn || ((P == Knight || P == Lance) && PromotionZone<2>(C, pos.checker_))) return;
            }
            *moves++ = Move::makeMove<NonPromote, true>(from, pos.checker_, PFrom, cap);
        });
    });
    //ʂ𓮂(checker͏)
    Bitboard toBB = attacks_bb(C, pos.king_[C], King, pos.occupied_) & ~pos.byColorBB_[C] & ~pos.checkerAttacksBB_;
    for_each_bb(toBB, [&](const Square to) {
        if (pos.checker_ == to) return;
        *moves++ = Move::makeMove<NonPromote, true>(pos.king_[C], to, King, pos.piece_[to]);
    });

    return moves;
}

template <Color C>
Move *generate_evasions_interpose(Move *moves, const Position& pos) {
    //
    if (pos.checked_ != 1) return moves;//̂Ƃ͍łȂ

    const int step = Direction[pos.king_[C]][pos.checker_];
    const Square to_first = pos.king_[C] + step;
    if (step == 0 || to_first == pos.checker_) return moves;

    uint8_t attacks[SquareNum];
    Bitboard pawnBB;
    if (pos.hand_[C][Pawn] > 0) pawnBB = pos.byTypeBB_[Pawn] & pos.byColorBB_[C];
    //1Tڂ͗̂ꏊցA2Tڂ͗̂Ȃꏊ֍
    for (int i = 0; i < 2; ++i) {
        for (Square to = to_first; to != pos.checker_; to += step) {
            if (i == 0) {
                attacks[to] = /*(to == to_first) || */!pos.attackersTo<C, true>(to, pos.occupied_).isNone();
                if (!attacks[to]) continue;
            } else {
                if (attacks[to]) continue;
            }
            for_each_piece<true>([&](auto o) {
                //(ړ)
                Bitboard fromBB = attacks_bb(OppositeColor(C), to, P, pos.occupied_) & pos.byTypeBB_[P] & pos.byColorBB_[C];
                for_each_bb(fromBB, [&](const Square from) {
                    if (CanPromote(P) && (PromotionZone(C, to) || PromotionZone(C, from))) {
                        *moves++ = Move::makeMove<Promote>(from, to, PFrom);
                        if (P == Rook || P == Bishop || P == Pawn || ((P == Knight || P == Lance) && PromotionZone<2>(C, to))) return;
                    }
                    *moves++ = Move::makeMove<NonPromote>(from, to, PFrom);
                });

                //()
                if (P >= HandTypeNum) return;
                if (pos.hand_[C][P] == 0) return;

                if (P == Pawn && !pawn_dropable(to, pawnBB)) return;
                if ((P == Pawn || P == Lance) && PromotionZone<1>(C, to)) return;
                if (P == Knight && PromotionZone<2>(C, to)) return;
                *moves++ = Move::makeDrop(P, to);
            });
        }
    }
    return moves;
}

template <Color C>
Move *generate_evasions(Move *moves, const Position& pos) {
    moves = generate_evasions_base<C>(moves, pos);
    moves = generate_evasions_interpose<C>(moves, pos);
    return moves;
}

template <Color C>
Move *generate_evasions_ex(Move *moves, const Position& pos) {
    //p̕sA2iڂ̍̕s
    if (pos.checked_ != 1) return moves;
    const PieceType cap = pos.piece_[pos.checker_];
    //checker
    for_each_piece([&](auto o) {
        if (P != Rook && P != Bishop && P != Lance && P != Pawn) return;
        if (P == Pawn && PromotionZone<1>(C, pos.checker_)) return;
        if (P == Lance && !RankBB[RankInv(C, 1)].isSet(pos.checker_)) return;

        Bitboard fromBB = attacks_bb(OppositeColor(C), pos.checker_, P, pos.occupied_) & pos.byTypeBB_[P] & pos.byColorBB_[C];
        for_each_bb(fromBB, [&](const Square from) {
            if (PromotionZone(C, pos.checker_) || PromotionZone(C, from)) {
                *moves++ = Move::makeMove<NonPromote>(from, pos.checker_, P, cap);
            }
        });
    });
    //ړ
    const int step = Direction[pos.king_[C]][pos.checker_];
    if (step != 0) {
        for (Square to = pos.king_[C] + step; to != pos.checker_; to += step) {
            for_each_piece<true>([&](auto o) {
                if (P != Rook && P != Bishop && P != Lance && P != Pawn) return;
                if (P == Pawn && PromotionZone<1>(C, to)) return;
                if (P == Lance && !RankBB[RankInv(C, 1)].isSet(to)) return;

                Bitboard fromBB = attacks_bb(OppositeColor(C), to, P, pos.occupied_) & pos.byTypeBB_[P] & pos.byColorBB_[C];
                for_each_bb(fromBB, [&](const Square from) {
                    if (PromotionZone(C, to) || PromotionZone(C, from)) {
                        *moves++ = Move::makeMove<NonPromote>(from, to, P);
                    }
                });
            });
        }
    }
    return moves;
}

template <Color C>
Move *generate_pseudo_legal(Move *moves, const Position& pos) {
    if (pos.checked_) {
        moves = generate_evasions<C>(moves, pos);
        moves = generate_evasions_ex<C>(moves, pos);
    } else {
        moves = generate_non_evasions<C>(moves, pos);
        moves = generate_non_evasions_ex<C>(moves, pos);
    }
    return moves;
}

Move *generate_pseudo_legal(Move *moves, const Color c, const Position& pos) {
    return (c == Black) ? generate_pseudo_legal<Black>(moves, pos) : generate_pseudo_legal<White>(moves, pos);
}
