
#include <iomanip>
#include <sstream>
#include <random>
#include <chrono>
#include <thread>

#include "search.h"
#include "thread.h"
#include "movegen.h"
#include "tt.h"
#include "usi.h"

#define ENABLE_TSUMERO
#define ENABLE_RAZOR
#define PRUNING true

//Tn߂ƏI鎞
std::chrono::high_resolution_clock::time_point time_start, time_end;
int t_current;//Ȃꍇ̗\vl
thread_local int pvcount;
std::atomic<bool> in_ponder = false;
static std::mutex mutex;

Score mate_score;

long long last_time;
Move prev_best;

LEARN_THREAD_LOCAL History history;
LEARN_THREAD_LOCAL Gains gains;
LEARN_THREAD_LOCAL Move counter[SquareNum][ColorNum][PieceTypeNum][2];
LEARN_THREAD_LOCAL Move followup[SquareNum][ColorNum][PieceTypeNum][2];

forceinline Move *countermoves(const Color c, const Move move) {
    return move.isNone() ? nullptr : counter[move.to()][c][move.piece()];
}
forceinline Move *followupmoves(const Color c, const Move move) {
    return move.isNone() ? nullptr : followup[move.to()][c][move.piece()];
}

void RootMove::extractPvFromTT(Position& pos) {
    Position pos_ = pos;
    Color c = pos.root_turn_;
    if (pos_.ply_ % 2) c = OppositeColor(c);
    ply_ = 0;
    assert(!pv_[0].isNone());
    assert(pos_.moveIsPseudoLegal(c, pv_[0]));

    std::lock_guard<std::mutex> lock{ mutex };

    for (int ply = 0; ply < MaxPly - 3; ++ply) {
        TTEntry entry;
        Move move;

        if (ply == 0) {
            move = pv_[0];
            pv_[ply + 1] = Move::None();
            assert(!move.isNone());
        } else {
            if (!tt.probe(entry, pos_.key_)) {
                if (ply_ == 0) ply_ = ply;
                break;
            }
            if (ply_ == 0 && entry.depth() <= DepthQuiet) ply_ = ply;
            move = entry.move();
            //`FbNĂȂ
        }
        if (move.isNone() || !pos_.moveIsPseudoLegal(c, move)) {
            if (ply_ == 0) ply_ = ply;
            assert(ply != 0);
            break;
        }
        pos_.doMoveInPlace(c, move);
        Bitboard checkersBB = (c == Black) ? pos_.attackersTo<White>(pos_.king_[Black], pos_.occupied_) : pos_.attackersTo<Black>(pos_.king_[White], pos_.occupied_);
        pos_.checked_ = static_cast<uint8_t>(checkersBB.popcnt());
        c = OppositeColor(c);
        pv_[ply] = move;
        pv_[ply + 1] = Move::None();
    }
}
void RootMove::insertPvInTT(Position& pos) {
    Color c = (Color)(((int)pos.root_turn_ + pos.ply_) % 2);
    Position pos_ = pos;

    for (int ply = 0; ply < MaxPly - 3; ++ply) {
        TTEntry entry;
        Move move = pv_[ply];

        if (move.isNone() || !pos_.moveIsPseudoLegal(c, move)) break;
        if (!tt.probe(entry, pos_.key_)) {
            tt.store(pos_.key_, ScoreZero, BoundNone, DepthZero, move);
        }

        pos_.doMoveInPlace(c, move);
        Bitboard checkersBB = (c == Black) ? pos_.attackersTo<White>(pos_.king_[Black], pos_.occupied_) : pos_.attackersTo<Black>(pos_.king_[White], pos_.occupied_);
        pos_.checked_ = static_cast<uint8_t>(checkersBB.popcnt());
        c = OppositeColor(c);
    }
}

std::atomic<bool> stop;
LEARN_THREAD_LOCAL uint64_t nodes[8];

std::vector<BookEntry> v_book;

int16_t Reductions[2][64][64];//[pv][depth][moveNumber]
forceinline Depth reduction(const bool PVNode, const Depth depth, const int move_count) {
    return static_cast<Depth>(Reductions[PVNode][std::min(DepthToPly(depth), 63)][std::min(move_count, 63)]);
}
Score FutilityMargins[16][64];//[depth][moveCount]
int FutilityMoveCounts[32];//[depth]

void init_search_table() {
    int d;  // depth (ONE_PLY == 2)
    int hd; // half depth (ONE_PLY == 1)
    int mc; // moveCount

    // Init reductions array
    for (hd = 1; hd < 64; hd++) {
        for (mc = 1; mc < 64; mc++) {
            double    pvRed = log(double(hd)) * log(double(mc)) / 3.0;
            double nonPVRed = 0.33 + log(double(hd)) * log(double(mc)) / 2.25;
            Reductions[1][hd][mc] = (int16_t)(pvRed >= 1.0 ? floor(pvRed * int(OnePly)) : 0);
            Reductions[0][hd][mc] = (int16_t)(nonPVRed >= 1.0 ? floor(nonPVRed * int(OnePly)) : 0);
        }
    }

    for (d = 1; d < 16; ++d) {
        //unsigned long m;
        //_BitScanReverse(&m, d * d);
        for (mc = 0; mc < _countof(*FutilityMargins); ++mc) {
            FutilityMargins[d][mc] = static_cast<Score>(112 * static_cast<int>(log(static_cast<double>(d*d) / 2) / log(2.0) + 1.001) - 8 * mc + 45);
        }
    }

    // init futility move counts
    for (d = 0; d < 32; ++d) {
        FutilityMoveCounts[d] = static_cast<int>(3.001 + 0.3 * pow(static_cast<double>(d), 1.8));
    }
}


forceinline bool is_quiet(Move move) {
    return !move.isCapture() && (!move.isPromotion() || move.piece() == Silver);
}

forceinline constexpr Score MateIn(const int ply) {
    return ScoreMate0Ply - static_cast<Score>(ply);
}

//ǂ݋؂Ȃǂ̏GUIɑ
void info_to_USI(Position& pos, const Score alpha, const Score beta, const Score score, const Depth depth, const RootMove rm, const bool done = false) {
    const auto duration = std::chrono::high_resolution_clock::now() - time_start;//o߂
    const auto msec = std::chrono::duration_cast<std::chrono::milliseconds>(duration).count();//o߂(~bP)
    const auto msec2 = std::max(msec, static_cast<decltype(msec)>(1));
    const bool A = false;//SmartPvOutput;

    std::lock_guard<std::mutex> lock{ mutex };

    assert(!rm.pv_[0].isNone());

    if (SmartPvOutput) {
        if (depth < PlyToDepth(2) && abs(score) < ScoreMateInMaxPly) return;
        if (msec - last_time < (last_time == 0 ? 60 : 300) && abs(score) < ScoreMateInMaxPly && rm.pv_[0] == prev_best) return;
    }
    if (score == -ScoreInfinite) return;
    if (abs(score) > abs(mate_score)) {
        if (abs(score) >= ScoreMateInMaxPly) mate_score = score;
    } else {
        if (abs(mate_score) > ScoreMateInMaxPly/* && !(alpha < score && score < beta)*/) return;
    }
    last_time = msec;
    prev_best = rm.pv_[0];

    if (!A && done) {
        std::cout << "info" << " time " << msec << " nodes " << nodes[0]
            << " nps " << nodes[0] * 1000 / msec2 << " string "
            << "[" << pos.this_thread_->idx_ << "] "
            << ((score <= alpha) ? "" : (score >= beta) ? "" : "")
            << " (" << std::setw(5) << alpha << "," << std::setw(5) << beta << ")"
#ifdef MATE_TEST
            << " n[1]" << nodes[1] << " n[2]" << nodes[2] << " n[3]" << nodes[3] << " n[4]" << nodes[4]
#endif
            << std::endl;

    } else if (!A || (done && ((score > alpha && msec > 30) || abs(score) >= ScoreMateInMaxPly))) {
        std::stringstream ssScore;
        if (abs(score) >= ScoreMateInMaxPly && abs(score) < ScoreMate0Ply) {
            if (score > 0) {
                ssScore << "mate +" << ScoreMate0Ply - score;
            } else {
                ssScore << "mate -" << ScoreMate0Ply + score;
            }
        } else {
            ssScore << "cp " << score;
        }
        //if (A) {
            if (score <= alpha) ssScore << " upperbound";
            if (score >= beta) ssScore << " lowerbound";
        //}

        std::string pv;
        for (int ply = 0; ply < MaxPly; ++ply) {
            Move move = rm.pv_[ply];
            if (move.isNone()) break;
            pv += ' ' + move.toUSI();
        }
        pv = pv.empty() ? "string [resign]" : "pv" + pv;

        std::cout << "info" << " depth " << DepthToPly(depth) << " seldepth " << rm.ply_ << " time " << msec << " nodes " << nodes[0]
            << " nps " << nodes[0] * 1000 / msec2 << " score " << ssScore.str() << " " << pv << std::endl;
    }
#ifdef MATE_TEST
    std::cout << "info string " << nodes[1] << " " << nodes[2] << " " << nodes[3] << " " << nodes[4] << std::endl;
#endif
}

void update_stats(const Color c, Position& pos, const Move move, const Score score, const Depth depth, const Move *quiets, const size_t quiets_count) {
    assert(move == move.noScore());
    assert(!move.isNone());
    if (pos.checked_) return;
    if (!is_quiet(move)) return;

    Thread *thisThread = pos.this_thread_;

    if (is_quiet(move)) {
        assert(move == move.noScoreQuiet());
        int d = DepthToHalfPly(depth);
        Score bonus = static_cast<Score>(d * d);//
        history.update(c, move, bonus);
        if (quiets) {
            for (int i = 0; i < quiets_count - 1; ++i) {//JbgNŌ̎(vmove)͏
                history.update(c, quiets[i], -bonus);
            }
            //assert(quiets[quiets_count - 1].noScoreQuiet() == move);//SL^ꍇ̂
        }
    }

    if (move != pos.killer_move_[0]) {
        pos.killer_move_[1] = pos.killer_move_[0];
        pos.killer_move_[0] = move;
    }

    Move *moves;
    moves = countermoves(c, pos.previous_move_);
    if (moves && move != moves[0]) {
        moves[1] = moves[0];
        moves[0] = move;
    }

    moves = followupmoves(c, pos.forward(-1).previous_move_);
    if (moves && move != moves[0]) {
        moves[1] = moves[0];
        moves[0] = move;
    }
}

//Î~T܂
template <Color C, bool Learn>
Score search(Position& pos, const Score original_alpha, Score beta, const Depth depth, const bool cut_node) {
    //GtemplateĂȂȂ̂ŃRpCxD
    const bool Quiescence = (depth <= DepthQuiet);
    const bool InCheck = (pos.checked_ != 0);
    const bool RootNode = (pos.ply_ == 0);
    const bool PVNode = (beta - original_alpha > 1);

    //u\nbVƂ𓾂
    TTEntry entry;
    bool entry_available = false;
    bool in_tsumero = false;//̋ǖʂl߂납
    Move tt_move = Move::None();
    if (tt.probe(entry, pos.key_)) {
        if (pos.ply_ == 5 && entry.ply1() && entry.generation() == tt.generation()) return ScoreMateInMaxPly;//ʑ΍
        if (/*!PVNode && */!RootNode && entry.depth() >= depth) {
            Bound bound = entry.bound();
            Score score = entry.score();
            if (abs(score) > ScoreMateInMaxPly) score -= pos.ply_ * Sign(score);
            if ((PVNode && bound == BoundExact) || (!PVNode && (
                ((bound & BoundUpper) && score < beta) ||//PVł͂Ȃ̂beta菬alphaȉƂȂfail-low
                ((bound & BoundLower) && score >= beta)))) {
                tt_move = entry.move();
                if (!Quiescence && score >= beta && !tt_move.isNone()) {
                    update_stats(C, pos, tt_move, score, depth, nullptr, 0);
                }
                pos.next().previous_move_ = tt_move;
                return score;
            }
        }
        entry_available = true;
        tt_move = entry.move();
        if (!tt_move.isNone() && !pos.moveIsPseudoLegal<C>(tt_move)) {
#ifdef DEBUG
            std::cout << pos.toKifu() << " : " << tt_move.toUSI() << std::endl;
            assert(0);
#endif
            std::cout << "info string hash collision." << pos.toKifu() << std::endl;
            tt_move = Move::None();
            entry_available = false;
        }
    }

    if (!Learn && !RootNode) {
        //Ƃo
        const int k = Quiescence ? 4 : (depth < PlyToDepth(3)) ? 8 : 20;//k萔
        for (int i = 4; i <= k; i += 2) {
            if (pos.forward(-i).key_ == 0) break;

            //(board_key_1 == board_key_2 肷) key1 - key2 == hand64_2 - hand64_1;
            const int64_t diff = pos.key_ - pos.forward(-i).key_;
            if (diff == 0) {
                //
                if (pos.continuous_check_[C] >= i) return -ScoreMate0Ply;
                if (pos.continuous_check_[OppositeColor(C)] >= i) return ScoreMate0Ply;
                return static_cast<Score>(-70) * ColorToSign(C);
            }

            //board_key_Ƃ͕ۏ؂łȂ̂ŃXN(ł郊XNƎv)
            constexpr int64_t Mask = 0x0101010101010100 * 3;//Position::hand64_Ŋe1̒l
            const int64_t neg_diff = -diff;
            //-b == ~b + 1; ~b == -b - 1; popcnt(b) <= 1  (b & (b - 1)) == 0
            //if ((diff & ~Mask) == 0 && (diff & ~neg_diff) == 0) {
            if ((diff & ~Mask) == 0) {
                return ScoreMate0Ply;//̎
            }
            //if ((neg_diff & ~Mask) == 0 && (neg_diff & ~diff) == 0) {
            if ((neg_diff & ~Mask) == 0) {
                return -ScoreMate0Ply;//̎
            }
        }
    }

    Score alpha = original_alpha;
    if (!Learn && depth > PlyToDepth(5) && !RootNode) {
        for (int i = 5; i <= 20; i += 2) {
            if (pos.forward(-i).key_ == 0) break;

            const int64_t diff = pos.key_ - pos.forward(-i).key_;

            if (abs(diff) == 1) {
                return ScoreMateInMaxPly;
            }
        }
    }

    if (!RootNode) {
        // mate distance pruning
        alpha = std::max(-MateIn(pos.ply_), alpha);
        beta = std::min(MateIn(pos.ply_ + 1), beta);
        if (beta <= alpha) return alpha;
    }

    ++nodes[0];//Tm[h

    //1l ̓IȎ肪ԂȂ̂Root͏
    if ((Learn || !RootNode) && !InCheck && !entry_available) {
        const bool b1 = pos.isMateIn1Ply<C>();
#ifdef MATE_TEST
        auto c0 = pos.cut_node_;
        const bool b2 = pos.isMateIn1Ply_Slow<C>();
        assert(pos.cut_node_ == c0);
        pos.cut_node_ = c0;
        if (b1 && b2) ++nodes[1];//l OK
        if (!b1 && b2) ++nodes[2];//l܂Ȃ Ȃقǂ
        if (b1 && !b2) ++nodes[3];//lނƌF ΂ɑʖ
        if (!b1 && !b2) ++nodes[4];//l܂Ȃ OK
        if (b1 != b2) {
            std::cout << pos.toUSI() << std::endl;
            assert(!b1 && b2);
            if (nodes[2] == 1) std::cout << "info string minogashi." << std::endl;
        }
#endif
        //isWinŖĂԂ̉ʂĂȂ_T(l݂ƈĂ10͐錾łȂƂ킩邵A1l߂2x)
        if (b1 || pos.isWin<C>()) {
            Score score = MateIn(pos.ply_ + 1);
            tt.store(pos.key_, score + pos.ply_ * Sign(score), BoundExact, DepthMax, Move::None());
            return score;
        }
    }

    //]l̍vZ͂܂Œx
    assert(!RootNode || pos.static_eval_ != ScoreInfinite);
    if (pos.static_eval_ == ScoreInfinite) pos.static_eval_ = evaluate<OppositeColor(C)>(pos);


    //firstsecond͓Ԃ̎w
    auto allows = [&](const Move first, const Move second) {
        //𓮂
        if (first.to() == second.from()) return true;
        //ǂꏊ֓
        if (first.from() == second.to()) return true;
        //ǂĂ܂œ悤ɂȂ
        if (!first.isDrop() && !second.isDrop() && BetweenBB[second.from()][second.to()].isSet(first.from())) return true;
        //\߂Ђt
        const Occupied occupied = second.isDrop() ? pos.occupied_ : pos.occupied_ ^ SquareOcc[second.from()];
        const Bitboard attacksBB = attacks_bb(OppositeColor(C), first.to(), CvtIfGolds(pos.piece_[first.to()]), occupied);
        if (attacksBB.isSet(second.to())) return true;
        //O̎gĊJ
        if (attacksBB.isSet(pos.king_[C])) return true;

        return false;
    };
    //firstsecond͓Ԃ̎w
    auto allows3 = [&](const Move first, const Move second) {
        if (first.isNone()) return false;
        ////𓮂
        //if (first.to() == second.from()) return true;
        ////ǂꏊ֓
        //if (first.from() == second.to()) return true;
        //ǂĂ܂œ悤ɂȂ
        //if (!first.isDrop() && !second.isDrop() && BetweenBB[second.from()][second.to()].isSet(first.from())) return true;
        ////\߂Ђt
        const Occupied occupied = second.isDrop() ? pos.occupied_ : pos.occupied_ ^ SquareOcc[second.from()];
        const Bitboard attacksBB = attacks_bb(OppositeColor(C), first.to(), CvtIfGolds(pos.piece_[first.to()]), occupied);
        if (attacksBB.isSet(second.to())) return true;
        //O̎gĊJ
        if (attacksBB.isSet(pos.king_[C])) return true;

        return false;
    };
    //firstsecond͈قȂԂ̎w
    auto allows2 = [&](const Move first, const Move second) {
        if (first.isNone()) return false;
        //ǂꏊ֓
        if (first.from() == second.to() && pos.seeSign<C>(second, 0)) return true;
        //їꏊ
        if (!first.isDrop() && !isNegative(pos.pinCheck<C>(first.from(), first.to(), second.to(), pos.occupied_ | SquareOcc[second.to()]))) return true;

        if (pos.next().checked_) {
            //󏡂ɂȂfirst.from()֗炷
            if (/*!first.isDrop() && */first.piece() == King && attacks_bb(C, second.to(), CvtIfGolds(second.pieceTo()), pos.occupied_).isSet(first.from())) return true;

            //if (pos.next().checked_ == 2) return true;
            //if (BetweenBB[pos.next().checker_][pos.king_[OppositeColor(C)]].isSet(first.from())) return true;
        }

        return false;
    };
    //firstsecond͎Ԃ̈قȂA firstsecondז悤Ȏ肩 first͎łȂ
    auto refutes = [&](const Move first, const Move second) {
        //ꂻ̂𓦂
        if (first.from() == second.to()) return true;
        //Ђt ₪Ŏ悤ƂĂ̂ɂ̋ɂЂtĂӖȂ
        if (second.isCapture() && (second.piece() == King || CapturePieceScore[second.cap()] <= CapturePieceScore[second.piece()])) {
            assert(!pos.byTypeBB_[Empty].isSet(second.to()));
            Occupied occupied = pos.occupied_ ^ SquareOcc[second.from()];
            if (attacks_bb(C, first.to(), CvtIfGolds(first.pieceTo()), occupied).isSet(second.to())) return true;

            if (!first.isDrop()) {
                Square pinner = pos.pinCheck<C>(first.from(), first.to(), second.to(), occupied);
                if (!isNegative(pinner)) return true;
            }
        }
        //їɎՂ
        if (!second.isDrop() && IsSlider(second.piece()) && BetweenBB[second.from()][second.to()].isSet(first.to()) && pos.seeSign<C>(first, 0)) {
            return true;
        }
        return false;
    };

#ifdef ENABLE_RAZOR
    bool razor_c = false;
    Score razor_score;
#endif
    Score best_score;
    Score futility_base;
    Move threat_move;
    Thread* thisThread = pos.this_thread_;

    if (Quiescence && !InCheck) {
        best_score = pos.static_eval_;
        if (!PVNode && entry_available && (entry.bound() & BoundLower) && entry.score() > best_score) best_score = entry.score();
        if (best_score >= beta) {
            tt.store(pos.key_, best_score, BoundLower, std::max(depth, DepthZero), Move::None());
            return best_score;
        }
        if (PVNode && best_score > alpha) alpha = best_score;
        futility_base = best_score + 128;

    } else {
        best_score = -ScoreInfinite;
    }

    if (!Quiescence) {
        threat_move = Move::None();
#ifdef ENABLE_TSUMERO
        if (!PVNode && entry_available) in_tsumero = entry.tsumero();
#endif
        pos.next().excluded_move_ = Move::None();
        pos.next().skip_nullmove_ = false;
        pos.next().reduction_ = DepthZero;
        std::memset(pos.forward(2).killer_move_, 0, sizeof Position::killer_move_);

        if (PRUNING && !pos.previous_move_.isNone() && !InCheck && !pos.forward(-1).checked_ && is_quiet(pos.previous_move_)) {
            //O̎w肪ÓI]lǂꂾグ(莋_-pos.static_eval_)
            gains.update(OppositeColor(C), pos.previous_move_, -pos.static_eval_ - pos.forward(-1).static_eval_);
        }

        if (PRUNING && !PVNode && !InCheck && !in_tsumero && abs(beta) < ScoreMateInMaxPly - 200) {
            Score eval = pos.static_eval_;
            if (entry_available && (entry.bound() & ((entry.score() > eval) ? BoundLower : BoundUpper))) eval = entry.score();

            //Razoring
            auto razor_margin = [](const Depth depth) {
                return static_cast<Score>(512 + 16 * DepthToHalfPly(depth));
            };
            if (depth < PlyToDepth(4) && eval + razor_margin(depth) < beta && tt_move.isNone()) {
                const Score rbeta = beta - razor_margin(depth);
                const Score s = search<C, Learn>(pos, rbeta - 1, rbeta, DepthQuiet, cut_node);
                if (s < rbeta) {
#ifdef ENABLE_RAZOR
                    razor_score = s;
                    razor_c = true;
#else
                    return s;
#endif
                }
            }

            if (!pos.skip_nullmove_ && abs(eval) < ScoreMateInMaxPly
#ifdef ENABLE_RAZOR
                && !razor_c
#endif
                ) {
                //Static null move pruning
                if (depth < PlyToDepth(4)) {
                    const Score s = add_if(eval - FutilityMargins[DepthToHalfPly(depth)][0], depth > PlyToDepth(3), Score(256));
                    if (s >= beta) return s;
                }

                //Null move
                if (depth >= PlyToDepth(2) && beta <= eval) {
                    Depth reduction = OnePly * 3 + (DepthToHalfPly(depth) / 4) * OnePly / 2;
                    if (beta < eval - PieceScore[Pawn]) reduction += OnePly;

                    assert(!pos.previous_move_.isNone());
                    assert(alpha + 1 == beta);
                    pos.doNullMove<C>();
                    Score nullScore = -search<OppositeColor(C), Learn>(pos.next(), -beta, -alpha, depth - reduction, !cut_node);
                    pos.next().skip_nullmove_ = 0;

                    if (beta <= nullScore) {
                        if (ScoreMateInMaxPly <= nullScore) nullScore = beta;

                        if (depth < PlyToDepth(6)) return nullScore;

                        assert(depth - reduction > PlyToDepth(0));
                        pos.skip_nullmove_ = 1;
                        const Score s = search<C, Learn>(pos, alpha, beta, depth - reduction, false);
                        pos.skip_nullmove_ = 0;

                        if (beta <= s) return nullScore;

                    } else {
#ifdef ENABLE_TSUMERO
                        //pX畉ɂȂ l߂낾ۂ
                        if (nullScore <= -ScoreMateInMaxPly) in_tsumero = true;
#endif
                        //肪Jbg ǂ̎ŃJbg
                        threat_move = pos.forward(2).previous_move_;
                        if (depth < PlyToDepth(5) && pos.reduction_ != 0 && !threat_move.isNone() && allows(pos.previous_move_, threat_move)) {
                            return beta - 1;//valpha em[hłReductionŃJbĝœǂݒƂɂȂ
                        }
                    }
                }

                //ProbCut
                if (depth >= PlyToDepth(5)) {
                    const Score rbeta = beta + 200;
                    const Depth rdepth = depth - 4 * OnePly;
                    const Score capture_threshold = CapturePieceScore[pos.previous_move_.cap()] + 1;

                    assert(PlyToDepth(1) <= rdepth);
                    assert(!pos.previous_move_.isNone());

                    GenerateMovePhase phase = ProbCut;
                    Move moves[MaxMove];
                    Move *current, *last;
                    while (1) {
                        Move move;
                        if (phase == ProbCut) {
                            ++phase;
                            move = tt_move;
                            if (move.isNone()) continue;
                        } else {
                            if (phase == PH_TacticalMoves1) {
                                ++phase;
                                last = generate_captures<C>(current = moves + 1, pos);
                                current[-1] = Move::MaxScore();
                                insertion_sort<true>(current, last, std::greater<>{});
                            }
                            if (current == last) break;
                            move = (current++)->noScore();
                            if (move == tt_move) continue;
                        }
                        if (!pos.pseudoLegalMoveIsLegal<C>(move)) continue;
                        if (!pos.seeSign<C>(move, capture_threshold)) continue;
                        pos.moveGivesCheck<C>(move);
                        pos.doMove(C, move);

                        Score score = -search<OppositeColor(C), Learn>(pos.next(), -rbeta, -rbeta + 1, rdepth, !cut_node);
                        pos.ex_move_ = Move::None();
                        if (score >= rbeta) return score;
                    }
                }
            }
        }
        //Internal iterative deepening
        if (depth >= (PVNode ? PlyToDepth(5) : PlyToDepth(9)) && tt_move.isNone() && (PVNode || (!InCheck && beta <= pos.static_eval_ + 256))) {
            const Depth d = PVNode ? depth - 2 * OnePly : HalfPlyToDepth(DepthToHalfPly(depth) / 2);

            pos.skip_nullmove_ = 1;
            search<C, Learn>(pos, alpha, beta, d, true);
            pos.skip_nullmove_ = 0;

            entry_available = tt.probe(entry, pos.key_);
            if (entry_available) tt_move = entry.move();
        }
    }

    const Move excluded_move = pos.excluded_move_;
    bool singular_extension_node = !RootNode && depth >= PlyToDepth(8) && !tt_move.isNone()
        && excluded_move.isNone() && (entry.bound() & BoundLower) && depth - 3 * OnePly <= entry.depth();

    const Depth tsumero_extension = (in_tsumero && pos.forward(-1).seeSign<OppositeColor(C)>(pos.previous_move_, 1)) ? OnePly / 2 : DepthZero;

    if (Quiescence && !excluded_move.isNone()) std::cout << excluded_move.toUSI() << std::endl;
    assert(!Quiescence || excluded_move.isNone());

    Move best_move = Move::None();
    bool best_move_is_check;
    int move_count = 0;
    int mated_count = 0;
    int quiets_count = 0;
    Move quiets_searched[128];

    Move moves[MaxMove + 2], killers[6 + 1];
    Move *const first = moves + 1;
    Move *current, *last, *bad_cap_last, *non_capture_first, *non_capture_last;
    GenerateMovePhase phase = InCheck ? EvasionSearch : !Quiescence ? MainSearch : (depth > DepthZero) ? QSearch : PH_QCaptures1;

    for (;; ++phase) {
        switch (phase) {
            case MainSearch: case QSearch: case EvasionSearch:
                if (tt_move.isNone()) continue;
                current = last - 1;//current͏ȂlastƓȂ last͏ĂȂw͎gȂ̂ŖȂ
                break;
            case PH_TacticalMoves0:
                non_capture_first = last = generate_captures<C>(current = bad_cap_last = first, pos);
                current[-1] = Move::MaxScore();
                insertion_sort<true>(current, last, std::greater<>{});
                break;
            case PH_Killers:
                current = last = killers;
                for (int i = 0; i < 2; ++i) {
                    Move move = pos.killer_move_[i];
                    if (!move.isNone()) *last++ = move;
                }
                Move *moves;
                moves = countermoves(C, pos.previous_move_);
                if (moves) {
                    for (int i = 0; i < 2; ++i) {
                        Move move = moves[i];
                        if (!move.isNone() && move != pos.killer_move_[0] && move != pos.killer_move_[1]) *last++ = move;
                    }
                }
                moves = followupmoves(C, pos.forward(-1).previous_move_);
                if (moves) {
                    for (int i = 0; i < 2; ++i) {
                        Move move = moves[i];
                        if (!move.isNone() && std::find(current, last, move) == last) *last++ = move;
                    }
                }
                *last = Move::None();
                break;
            case PH_NonTacticalMoves0:
                *non_capture_first++ = Move::MaxScore();//ԕ
                non_capture_last = generate_non_captures<C>(current = non_capture_first, pos);
                non_capture_last = generate_drops<C>(non_capture_last, pos);
                //XRÂ̂IŃ\[g
                last = std::partition(current, non_capture_last, [](const Move& m) { return m > Move::None(); });
                insertion_sort<true>(current, last, Move::greaterQuiet);
                break;
            case PH_NonTacticalMoves1:
                current = last;//partitioň둤
                last = non_capture_last;
                if (depth >= PlyToDepth(2)) std::sort(current, last, Move::greaterQuiet);//̃\[g̓RXg
                break;
            case PH_BadCaptures:
                current = first;//t MVV-LVAɂȂĂ
                last = bad_cap_last;
                break;
            case PH_Evasions0:
                last = generate_evasions_base<C>(current = first, pos);
                current[-1] = Move::MaxScore();
                insertion_sort<true>(current, last, std::greater<>{});
                break;
            case PH_Evasions1:
                last = generate_evasions_interpose<C>(current = first, pos);
                break;
            case PH_QCaptures0:
                last = generate_captures<C>(current = first, pos);
                current[-1] = Move::MaxScore();
                insertion_sort<true>(current, last, std::greater<>{});
                break;
            case PH_QCaptures1:
                last = generate_recaptures<C>(current = first, pos);
                break;
            case PH_Stop1: case PH_Stop2: case PH_Stop3: case PH_Stop4:
                goto break2;
            default:
                __assume(0);
        }

        while (current != last) {
            Move move;

            switch (phase) {
                case MainSearch: case QSearch: case EvasionSearch:
                    move = tt_move;
                    ++current;
                    break;
                case PH_TacticalMoves0://[first ][bad_cap_last ]non_capture_first
                    move = (current++)->noScore();
                    if (move == tt_move) {
                        continue;
                    } else if (!pos.seeSign<C>(move, 0)) {
                        *bad_cap_last++ = move;//noScoreȂ̂
                        continue;
                    }
                    break;
                case PH_Killers:
                    move = *current++;
                    if (move == tt_move || !pos.moveIsPseudoLegal<C>(move)) continue;
                    break;
                case PH_NonTacticalMoves0:
                case PH_NonTacticalMoves1:
                    move = (current++)->noScoreQuiet();
                    if (move == tt_move) continue;
                    for (int i = 0; !killers[i].isNone(); ++i) {
                        if (move == killers[i]) goto continue2;
                    }
                    break;
                case PH_BadCaptures:
                    move = *current++;
                    if (move == tt_move) continue;
                    break;
                case PH_Evasions1:
                    move = *current++;
                    if (move == tt_move) continue;
                    break;
                case PH_Evasions0:
                case PH_QCaptures0:
                    move = (current++)->noScore();
                    if (move == tt_move) continue;
                    break;
                case PH_QCaptures1:
                    move = *current++;
                    break;
                default:
                    __assume(0);
            }
        move_pick_done:
            
            if (move == excluded_move) continue;

            const bool gives_check = pos.moveGivesCheck<C>(move);
#ifdef ENABLE_RAZOR
            if (razor_c && !(allows3(pos.forward(-1).previous_move_, move) || allows2(pos.previous_move_, move))) continue;
#endif
            Depth new_depth = depth - OnePly;

            //O} ܂ủ\
            if (Quiescence) {
                if (PRUNING && !PVNode && move != tt_move) {
                    if (!InCheck && !gives_check) {
                        Score futility_score = futility_base + CapturePieceScore[pos.piece_[move.to()]];
                        if (move.isPromotion()) {
                            futility_score += PromotePieceScore[move.piece()];
                        }
                        if (futility_score < beta) {
                            best_score = std::max(best_score, futility_score);
                            continue;
                        }
                        if (futility_base < beta && depth <= DepthQuiet - OnePly && !pos.seeSign<C>(move, 1)) {
                            best_score = std::max(best_score, futility_base);
                            continue;
                        }
                    }
                    if ((!InCheck || (best_score > -ScoreMateInMaxPly && !move.isCapture()))
                        && (!move.isPromotion() || move.piece() == Silver) && !pos.seeSign<C>(move, 1)) continue;
                }

            } else {
                ++move_count;

                if (PRUNING && !PVNode && is_quiet(move) && !InCheck && !gives_check && !in_tsumero && best_score > -ScoreMateInMaxPly) {
                    assert(move != tt_move && !move.isCapture());

                    // move count based pruning
                    if (depth < PlyToDepth(16) && FutilityMoveCounts[DepthToHalfPly(depth)] <= move_count
                        && (threat_move.isNone() || !refutes(move, threat_move))
                        && (!pos.seeSign<C>(move, 0) || (attacks_bb(C, move.to(), CvtIfGolds(move.pieceTo()), pos.occupied_) & pos.byColorBB_[OppositeColor(C)] & (pos.byTypeBB_[Rook] | pos.byTypeBB_[Dragon] | pos.byTypeBB_[Bishop] | pos.byTypeBB_[Horse])).isNone())
                        ) {
                        continue;
                    }

                    // score based pruning
                    const Depth predictedDepth = new_depth - reduction(PVNode, depth, move_count);
                    // gain  2{ɂB
                    const Score futilityScore = pos.static_eval_ +
                        ((predictedDepth < PlyToDepth(7)) ? FutilityMargins[DepthToHalfPly(std::max(predictedDepth, DepthQuiet))][std::min(move_count, 63)] : ScoreInfinite) +
                        2 * gains.value(C, move.isDrop(), move.piece(), move.to());

                    if (futilityScore < beta) {
                        best_score = std::max(best_score, futilityScore);
                        continue;
                    }

                    if (predictedDepth < PlyToDepth(4) && !pos.seeSign<C>(move, 0)) {
                        continue;
                    }
                }
            }

            //T Singular̂doMovȇO
            if (!Quiescence) {
                Depth extension = tsumero_extension;
                //艄
                if (gives_check) extension = OnePly / 2/* + OnePly / 2 * (thisThread->idx_ % 2)*/;

                // singuler extension
                const Score rBeta = entry.score() - DepthToHalfPly(depth);
                if (singular_extension_node && move == tt_move && extension < OnePly && abs(entry.score()) < ScoreMateInMaxPly) {
                    auto c0 = pos.next().checkerAttacksBB_;
                    auto c1 = pos.next().checked_;
                    auto c2 = pos.next().checker_;
                    pos.key_ ^= Position::ZobristExclusion;
                    pos.excluded_move_ = move;
                    pos.skip_nullmove_ = 1;
                    Score score = search<C, Learn>(pos, rBeta - 1, rBeta, HalfPlyToDepth(DepthToHalfPly(depth) / 2), cut_node);
                    pos.skip_nullmove_ = 0;
                    pos.excluded_move_ = Move::None();
                    pos.key_ ^= Position::ZobristExclusion;

                    if (score < rBeta) {
                        extension = OnePly;
                        if (rBeta >= beta) extension += OnePly / 2;
                    }
                    pos.next().checkerAttacksBB_ = c0;
                    pos.next().checked_ = c1;
                    pos.next().checker_ = c2;
                }

                new_depth += extension;
            }

            //}肪ςł獇@肩mF ςłH
            if (!pos.pseudoLegalMoveIsLegal<C>(move)) { --move_count; continue; }
            //TquietsۑĂ
            if (!Quiescence && is_quiet(move) && quiets_count < _countof(quiets_searched)) {
                quiets_searched[quiets_count++] = move;
            }
            //i߂ pos.next()g
            pos.doMove(C, move);

#ifdef DEBUG
            uint64_t c0 = pos.next().attackersTo<OppositeColor(C)>(pos.next().king_[C], pos.next().occupied_).popcnt();
            uint64_t c1 = pos.next().attackersTo<C>(pos.next().king_[OppositeColor(C)], pos.next().occupied_).popcnt();
            uint64_t c2 = pos.next().checked_;
            if (c0 != 0 || c1 != c2) {
                //u܂͉̔F~X
                std::cout << pos.next().toKifu() << std::endl;
                std::cout << "info string " << c0 << " " << c1 << " " << c2 << std::endl;
                assert(0);
            }
#endif

            //̐[T
            Score score;
            assert(pos.next().reduction_ == 0);
            if (Quiescence) {
                score = -search<OppositeColor(C), Learn>(pos.next(), -beta, -alpha, new_depth, false);

            } else {
                const bool isPVMove = (PVNode && move_count == 1);//PVNodeœǂލŏ̎ nbV肪ȂƂ
                if (!isPVMove) {
                    bool lmr = (PRUNING && depth >= PlyToDepth(3) && phase != PH_Killers && is_quiet(move) && move != tt_move);
                    score = ScoreInfinite;//LMRȂꍇ
                    if (lmr && (pos.next().reduction_ = add_if(reduction(PVNode, depth, move_count), !PVNode && cut_node, OnePly)) != 0) {
                        //LMR
                        score = -search<OppositeColor(C), Learn>(pos.next(), -(alpha + 1), -alpha,
                                                                 std::max(new_depth - static_cast<Depth>(pos.next().reduction_), PlyToDepth(1)), true);
                        pos.next().reduction_ = 0;
                    }
                    assert(alpha < beta);
                    if (score > alpha) {
                        //PVNodeScoutƁANonPV̒ʏ̒TƁALMRŃ𒴂Ƃ̍ĒT
                        score = -search<OppositeColor(C), Learn>(pos.next(), -(alpha + 1), -alpha, new_depth, lmr || !cut_node);
                    }
                }

                //PVNodê PVMoveAScoutŃ𒴂Ƃ RootNodeȊOFullDepthSearchŃȏȂĒTɃJbg
                if (PVNode && (isPVMove || (alpha < score && (RootNode || score < beta)))) {
                    score = -search<OppositeColor(C), Learn>(pos.next(), -beta, -alpha, new_depth, false);
                }
            }

            //߂Kv͂Ȃ

            //vlf
            if (!Learn && (stop || std::chrono::high_resolution_clock::now() >= time_end)) {
                ::stop = true;
                return ScoreZero;//stoptruêƂ̓XRAgȂ̂œKɕԂ
            }

            if (!Quiescence && score < -ScoreMateInMaxPly) {
                ++mated_count;
            }

            if (RootNode) {
                auto it = std::find(thisThread->root_moves_.begin(), thisThread->root_moves_.end(), move);
                assert(it != thisThread->root_moves_.end());
                RootMove& rm = *it;
                assert(rm.pv_[0] == move);
                if (Learn) {
                    rm.score_ = (move_count == 1 || score > alpha) ? score : -ScoreInfinite;
                    rm.score_for_learn_ = score;
                    rm.extractPvFromTT(pos);
                }
                if (move_count == 1 || score > alpha) {
                    rm.score_ = score;
                    rm.extractPvFromTT(pos);

                    if (move_count > 1) ++pvcount;
                    //if (score > alpha) thisThread->pv_idx_++;

                    if (!Learn) {
                        //̃XbhbestmoveXV
                        //if (score > alpha) {
                            auto best_rm = std::max_element(thisThread->root_moves_.begin(), thisThread->root_moves_.end());
                            thisThread->best_score_ = best_rm->score_;
                            thisThread->best_move_ = best_rm->pv_[0];
                        //}
                        ////悳ȃXbhI
                        //Thread *best_thread = &threads[0];
                        //for (int i = 1; i < thread_num; ++i) {
                        //    if (threads[i].best_score_ > best_thread->best_score_ &&
                        //        (threads[i].completed_depth_ > best_thread->completed_depth_ || threads[i].best_score_ >= ScoreMateInMaxPly)) {
                        //        best_thread = &threads[i];
                        //    }
                        //}
                        //̂ǂ̃XbhXRA
                        bool is_best_score = true;
                        for (int i = 0; i < thread_num; ++i) {
                            if (i == thisThread->idx_) continue;
                            if (threads[i].best_score_ >= score + (int)(thisThread->idx_ == 0)) is_best_score = false;
                        }
                        //ǂ݋؂\
                        if (is_best_score && score > -ScoreMateInMaxPly) {
                        //if (best_thread == thisThread) {
                            info_to_USI(pos, original_alpha, beta, score, depth, rm);
                            thisThread->pv_idx_++;
                            //std::cout << "info string " << original_alpha << " " << beta << std::endl;
                        }
                    }

                } else {
                    rm.score_ = -ScoreInfinite;

                    if (!Learn) {
                        auto best_rm = std::max_element(thisThread->root_moves_.begin(), thisThread->root_moves_.end());
                        thisThread->best_score_ = best_rm->score_;
                        thisThread->best_move_ = best_rm->pv_[0];
                    }
                }
            }

            if (score > best_score) {
                best_score = score;
                best_move = move;
                best_move_is_check = gives_check;
                if (pos.ex_move_.isNone()) {
                    if (score >= beta) goto break3;
                    if (score > alpha) alpha = score;
                }
            }

            if (!pos.ex_move_.isNone()) {
                //łl߂̂ŕsǂł݂
                assert(move == pos.ex_move_.togglePromotion());

                RootMove rm0 = RootMove{ move };
                RootMove rm1 = RootMove{ pos.ex_move_ };
                rm0.extractPvFromTT(pos);
                std::memcpy(rm1.pv_, rm0.pv_, sizeof rm0.pv_);
                rm1.pv_[0] = pos.ex_move_;
                rm1.insertPvInTT(pos);

                move = pos.ex_move_;
                pos.ex_move_ = Move::None();
                --move_count;

                goto move_pick_done;
            }

        continue2:
            assert(1);//_~[
        }
    }
break2:

#ifdef ENABLE_RAZOR
    if (razor_c && best_score < beta) return razor_score;
#endif
    if (best_score == -ScoreInfinite) {
        if (excluded_move.isNone()) {
            if (pos.previous_move_.from() == PawnDrop && InCheck) {
                return ScoreInfinite;//łl
            } else {
                return -MateIn(pos.ply_);
            }
        } else {
            best_score = original_alpha;
        }
    }
break3:

    if (!excluded_move.isNone()) return best_score;
    if (!Quiescence && best_score >= beta && phase != PH_TacticalMoves0 && !best_move.isNone()) {//łl߂ɂ鏟
        update_stats(C, pos, best_move, best_score, depth, quiets_searched, quiets_count);
    }
    
    //u\ɏ
    const Bound bound = (best_score >= beta) ? BoundLower : (best_score > original_alpha) ? BoundExact : BoundUpper;
    Score score = best_score;
    if (abs(score) > ScoreMateInMaxPly) score += pos.ply_ * Sign(score);
    if (!Quiescence && !in_tsumero && !InCheck && best_score > -ScoreMateInMaxPly && mated_count * 2 >= move_count/* && mated_count > 1*/) {
        if (move_count - mated_count <= 1 + move_count / 2) in_tsumero = true;
        //mated_count >= (move_count - 1) * 7 / 8
    }
    assert(!((int)in_tsumero & ~1));
    tt.store(pos.key_, score, bound, std::max(depth, DepthZero), best_move, in_tsumero, pos.ply_ == 1);

    return best_score;
}
template Score search<Black, true>(Position& pos, const Score original_alpha, const Score beta, const Depth depth, const bool cut_node);
template Score search<White, true>(Position& pos, const Score original_alpha, const Score beta, const Depth depth, const bool cut_node);

//VǖʂŒTn߂Ȍ
void new_search() {
    tt.incrementGeneration();

    history.clear();
    gains.clear();
    std::memset(counter, 0, sizeof counter);
    std::memset(followup, 0, sizeof followup);

    std::memset(nodes, 0, sizeof nodes);
    mate_score = ScoreZero;
    last_time = 0;
    prev_best = Move::None();
    ::stop = false;
    Thread::count_ = (uint32_t)MaskBits(0, thread_num);
}

void new_search_each_thread(Position& pos) {
    Thread *thisThread = pos.this_thread_;

    std::memset(pos.killer_move_, 0, sizeof pos.killer_move_);
    std::memset(pos.next().killer_move_, 0, sizeof pos.next().killer_move_);
    for (int i = 0; i < MaxPly; ++i) pos.forward(i).ex_move_ = Move::None();
    pos.static_eval_ = evaluate<ColorNum>(pos) * ColorToSign(pos.root_turn_);
    pos.ply_ = 0;

    Move moves[MaxMove];
    Move *last = generate_pseudo_legal(moves, pos.root_turn_, pos);
    thisThread->root_moves_.clear();
    std::for_each(moves, last, [&](const Move& move) {
        thisThread->root_moves_.push_back(RootMove{ move.noScore() });
    });
    thisThread->completed_depth_ = DepthZero;
    thisThread->root_moves_[0].score_ = -ScoreInfinite;
    thisThread->best_score_ = -ScoreInfinite;
    thisThread->best_move_ = Move::None();
    assert(pos.excluded_move_.isNone());
    assert(pos.reduction_ == 0);
    assert(pos.key_ == pos.key(pos.root_turn_));
}

//[
template <Color C>
void id_loop(Position& pos, const Ply max_ply) {
    Score score = ScoreZero;
    new_search_each_thread(pos);

    //MaxPlyɂ̂āATǂ납Î~TȂƂOɂĂ
    for (Depth d = PlyToDepth(1); d < PlyToDepth(max_ply); d += OnePly / 2) {
        if (d >= PlyToDepth(MaxPly - 32)) break;
        if (pos.this_thread_->idx_ != 0) {
            unsigned long msb;
            _BitScanReverse64(&msb, pos.this_thread_->idx_ + 1);
            int t = (d + pos.this_thread_->idx_) >> (msb - 1);
            if (t % 2 || (nodes[0] <= 2048 && t % 4)) continue;
        }
        pos.this_thread_->root_depth_ = d;

        Score w = (Score)40;
        Score alpha = score - w, beta = score + w;

        std::mt19937_64 mt;

        for (;;) {
            pvcount = 0;
            pos.this_thread_->pv_idx_ = 0;

            if (nodes[0] <= 4096 || abs(score) >= ScoreMateInMaxPly) {
                //ǂݐ؂ǁAƒZl݂T
                alpha = -ScoreInfinite;
                beta = ScoreInfinite;
            }
            alpha = std::max(alpha, -MateIn(2));
            beta = std::min(beta, MateIn(1));
            if (alpha >= beta) std::cout << alpha << " " << beta << std::endl;
            assert(alpha < beta);
            assert(pos.excluded_move_.isNone());
            if (max_ply != MaxPly) {
                score = search<C, true>(pos, alpha, beta, d, false);
            } else {
                score = search<C, false>(pos, alpha, beta, d, false);
            }

            auto& rms = pos.this_thread_->root_moves_;
            if (rms.size() > 1) std::stable_sort(rms.begin(), rms.end(), std::greater<>{});
            for (RootMove& rm : rms) {
               rm.insertPvInTT(pos);
            }

            if (::stop) break;
            if (max_ply != MaxPly && nodes[0] > 256 + mt() % 512) break;
            if ((pos.this_thread_->idx_ == 0 || score <= -ScoreMateInMaxPly) && score != -ScoreMate0Ply && pos.this_thread_->pv_idx_ == 0) {
                if (pos.this_thread_->best_move_.isNone()) {
                    pos.this_thread_->best_move_ = rms[0].pv_[0];
                    pos.this_thread_->best_score_ = rms[0].score_;
                }
                if (max_ply == MaxPly) info_to_USI(pos, alpha, beta, score, d, rms[0]);
                //std::cout << "info string fail-low" << std::endl;
            }

            if (abs(score) >= MateIn(2)) break;
            if (alpha < score && score < beta) break;

            if (score <= alpha) {
                alpha = score - w;
            } else {
                beta = score + w;
            }
            w = w * 3 / 2 + 40;

            if (!in_ponder && t_current != 0 && d > PlyToDepth(3) && pos.this_thread_->idx_ == 0) {
                t_current += t_min / 8;
                if (pvcount > 0) t_current += t_min / 4;
                t_current = std::max((t_current + 999) / 1000 * 1000, 1000) - Margin[0];
                t_current = std::min(t_current, t_max);
                time_end = time_start + std::chrono::milliseconds(t_current);
                std::cout << "t_current:" << t_current << " t_min:" << t_min << " t_max:" << t_max << std::endl;
            }
        }

        if (::stop) break;
        pos.this_thread_->completed_depth_ = d;

        if (abs(score) >= MateIn(2)) break;
        if (abs(score) > ScoreMateInMaxPly && d >= PlyToDepth(8)) break;//ƒZl݂邩Ȃ
    }

    //TI CXbhI瑼̃XbhɂIĂ炤
    if (max_ply != MaxPly) return;
    if (pos.this_thread_->idx_ == 0) ::stop = true;

    const int count = (Thread::count_ ^= 1 << pos.this_thread_->idx_);

    //ȉ͍̏Ō̃Xbĥ
    if (count != 0) return;

    //ǂ݂牽ɏI
    if (in_ponder) {
        in_ponder = false;
        std::cout << "ponder end." << std::endl;
        return;
    }

    //bestmove߂
    Thread *best_thread = &threads[0];
    for (int i = 1; i < thread_num; ++i) {
        //if (threads[i].best_score_ > best_thread->best_score_ &&
        //    (threads[i].completed_depth_ >= best_thread->completed_depth_ || threads[i].best_score_ >= ScoreMateInMaxPly || best_thread->best_score_ == -ScoreInfinite)) {
        //    best_thread = &threads[i];
        //}
        if (threads[i].best_move_.isNone()) continue;
        if ((threads[i].best_score_ > best_thread->best_score_ && threads[i].completed_depth_ > best_thread->completed_depth_) || best_thread->best_move_.isNone()) {
            best_thread = &threads[i];
        }
    }
    Move best_move = best_thread->best_move_;
    score = best_thread->best_score_;
    if (score == -ScoreInfinite) best_move = Move::None();

    if (Ponder) in_ponder = !best_move.isNone() && abs(score) < ScoreMateInMaxPly;//̋ǖʂĂOɃZbg邽߁AbestmoveԂO
    std::cout << "info currmove " << best_move.toUSI() << std::endl;//̌œǂݐ؂Ƃm点邽 obhmEnEۂ
    std::cout << "bestmove " << best_move.toUSI() << std::endl;

    //Ŏ̒Tn߂鏀ł
    if (in_ponder) {
        time_start = std::chrono::high_resolution_clock::now();
        time_end = time_start + std::chrono::milliseconds(86400 * 1000);

        std::cout << "ponder start." << std::endl;
        Position *ppos = pos.this_thread_->ppos_;
        ppos->moveGivesCheck<C>(best_move);
        ppos->doMove(C, best_move);
        ppos->next().root_turn_ = OppositeColor(C);
        for (int i = 0; i < thread_num; ++i) {
            threads[i].ppos_++;
        }

        _mm_mfence();
        new_search();
    }
}

void thread_search(Thread *thisThread) {
    while (1) {
        Position& pos = thisThread->pos_[800];
        int j = 4;
        for (int i = 0; i < 1024; ++i) {
            pos.forward(-i) = thisThread->ppos_[-i];
            if (pos.forward(-i).key_ == 0) --j;
            if (j <= 0) break;
        }
        pos.this_thread_ = thisThread;

        (pos.root_turn_ == Black) ? id_loop<Black>(pos, MaxPly) : id_loop<White>(pos, MaxPly);

        //̒T܂őҋ@
        while (!(Thread::count_ & 1 << thisThread->idx_)) std::this_thread::sleep_for(std::chrono::milliseconds(1));
        _mm_mfence();
    }
}

template <Color C>
void think2(Position& pos) {
    //ʐ錾@
    if (!pos.checked_ && pos.isWin<C>()) {
        std::cout << "info string nyugyoku win" << std::endl;
        std::cout << "bestmove win" << std::endl;
        return;
    }

    //
    if (v_book.size() > 0 && game_ply < BookPly) {
        BookEntry book_entry;
        book_entry.key = pos.key_;
        auto range = std::equal_range(v_book.begin(), v_book.end(), book_entry, BookEntry::lessKey);
        size_t num = range.second - range.first;

        if (num > 0) {
            constexpr uint32_t N = 6;
            uint32_t sum = 0;
            std::for_each(range.first, range.second, [&](const BookEntry& a1) {
                if (a1.count < N) return;
                sum += a1.count;
            });
            std::random_device rnd;
            std::uniform_int_distribution<uint32_t> distribution(0, sum - 1);
            uint32_t t = distribution(rnd);

            //reverse_iterator𓾂@Ȃ́H
            Move best_move = Move::None();
            for (auto it = range.second; --it >= range.first;) {
                if (it->count < N) continue;
                if (best_move.isNone()) {
                    sum -= it->count;
                    if (sum <= t) best_move = it->move;
                }
                std::cout << "info nodes " << it->count << " pv " << it->move.toUSI() << std::endl;
            }
            if (!best_move.isNone() && pos.moveIsPseudoLegal<C>(best_move) && pos.pseudoLegalMoveIsLegal<C>(best_move)) {
                std::cout << "bestmove " << best_move.toUSI() << std::endl;
                return;
            } else {
                assert(0);
            }
        }
    }
    v_book.clear();

    if (threads) {
        for (int i = 0; i < thread_num; ++i) {
            threads[i].ppos_ = &pos;
        }
        _mm_mfence();
        new_search();

    } else {
        threads = static_cast<Thread *>(std::malloc(thread_num * sizeof Thread));
        std::memset(threads, 0, thread_num * sizeof *threads);
        for (int i = 0; i < thread_num; ++i) {
            threads[i].idx_ = i;
            threads[i].ppos_ = &pos;
        }
        new_search();
        _mm_mfence();
        for (int i = 0; i < thread_num; ++i) {
            threads[i].thread_ = std::thread{ thread_search, &threads[i] };
        }
    }
}

void think(Position& pos, const int msec) {
    t_current = msec;
    time_start = std::chrono::high_resolution_clock::now();
    time_end = time_start + std::chrono::milliseconds((t_current == 0) ? 86400 * 1000 : t_current);

    (pos.root_turn_ == Black) ? think2<Black>(pos) : think2<White>(pos);
}

template <Color C>
void perft_sub(Position& pos, Ply ply) {
    Move moves[MaxMove];
    const Move *last = generate_pseudo_legal<C>(moves, pos);
    --ply;

    for (Move *current = moves; current != last; ++current) {
        Move move = *current;
        if (!pos.pseudoLegalMoveIsLegal<C>(move)) continue;
        constexpr Color Them = OppositeColor(C);
        if (move.from() == PawnDrop && move.to() == pos.king_[Them] + FileNum * ColorToSign(C)
            && pos.attackersTo<Them, true>(move.to(), pos.occupied_).isNone()
            && !canKingEscape<C>(pos, step_attacks_bb(Them, King, pos.king_[Them]) & ~pos.byColorBB_[Them], pos.occupied_)
            ) continue;

        if (ply == 0) {
            ++nodes[0];
        } else {
            pos.moveGivesCheck<C>(move);
            pos.doMove(C, move);
            perft_sub<OppositeColor(C)>(pos.next(), ply);
        }
    }
}

void perft(Position& pos) {
    time_start = std::chrono::high_resolution_clock::now();
    stop = false;

    std::thread thread([&]() {
        for (Ply ply = 1; ply < MaxPly; ++ply) {
            nodes[0] = 0;
            (pos.root_turn_ == Black) ? perft_sub<Black>(pos, ply) : perft_sub<White>(pos, ply);
            auto duration = std::chrono::high_resolution_clock::now() - time_start;//o߂
            auto msec = std::chrono::duration_cast<std::chrono::milliseconds>(duration).count();//o߂(~bP)
            std::cout << "info" << " depth " << ply << " time " << msec << " nodes " << nodes[0]
                << " nps " << nodes[0] * 1000 / std::max(msec, static_cast<decltype(msec)>(1)) << " string perft" << std::endl;
        }
    });
    thread.detach();
}
