
#pragma once

#include <vector>
#include <chrono>
#include <atomic>

#include "bitboard.h"
#include "position.h"
#include "learn.h"

extern std::atomic<bool> stop;
extern std::atomic<bool> in_ponder;
extern LEARN_THREAD_LOCAL uint64_t nodes[8];
extern std::chrono::high_resolution_clock::time_point time_start, time_end;

inline constexpr Ply DepthToPly(const Depth depth) {
    return static_cast<Ply>((depth - (DepthQuiet - OnePly + 1)) / OnePly);
}
inline constexpr Depth PlyToDepth(const Ply ply) {
    return static_cast<Depth>(ply) * OnePly + (DepthQuiet - OnePly + 1);
}
forceinline constexpr int DepthToHalfPly(const Depth depth) {
    return (depth - (DepthQuiet - OnePly + 1)) / (OnePly / 2);
}
forceinline constexpr Depth HalfPlyToDepth(const int half_ply) {
    return static_cast<Depth>(half_ply) * (OnePly / 2) + (DepthQuiet - OnePly + 1);
}

void init_search_table();

template <Color C, bool Learn>
Score search(Position& pos, const Score original_alpha, const Score beta, const Depth depth, const bool cut_node);

void new_search();
void new_search_each_thread(Position& pos);

template <Color C>
void id_loop(Position& pos, const Ply max_ply);

void think(Position& pos, const int msec);

void perft(Position& pos);

//
struct alignas(16) BookEntry {
    uint64_t key;
    Move move;
    mutable uint32_t count;//constȃCe[^g(_ȁH)

    //
    static bool lessKey(const BookEntry& a1, const BookEntry& a2) {
        return a1.key < a2.key;
    }
    bool operator<(const BookEntry& a1) const {
        if (key == a1.key) {
            return Move::lessValue(move, a1.move);
        } else {
            return key < a1.key;
        }
    }
    static bool lessKeyGreaterCount(const BookEntry& a1, const BookEntry& a2) {
        if (a1.key == a2.key) {
            if (a1.count == a2.count) {
                return Move::lessValue(a1.move, a2.move);//͕sv
            } else {
                return a1.count > a2.count;
            }
        } else {
            return a1.key < a2.key;
        }
    }
};

extern std::vector<BookEntry> v_book;

template <bool Gain>
class Stats {
    //2͋łǂ2ʂ̂
    int16_t table_[ColorNum][2][PieceTypeNum][SquareNum];

public:
    static constexpr int16_t MaxScore = (1 << (32 - 1 - S4)) - 2;

    void clear() {
        memset(table_, 0, sizeof table_);
    }
    Score value(const Color c, const bool drop, const PieceType p, const Square to) const {
        return static_cast<Score>(table_[c][drop][p][to]);
    }
    void update(const Color c, const Move move, const Score s) {
        auto& t = table_[c][move.isDrop()][move.piece()][move.to()];
        if (Gain) {
            t = std::max<int16_t>(t - 1, s);
        } else {
            t = (s < ScoreZero) ? std::max<int16_t>(t + s, -MaxScore) : std::min<int16_t>(t + s, MaxScore);
        }
    }
};

using History = Stats<false>;
using Gains = Stats<true>;

extern LEARN_THREAD_LOCAL History history;

struct RootMove {
    Move pv_[MaxPly - 3];
    Score score_, score_for_learn_;
    Ply ply_;

    void extractPvFromTT(Position& pos);
    void insertPvInTT(Position& pos);
    bool operator<(const RootMove& rm) const {
        return score_ < rm.score_;
    }
    bool operator>(const RootMove& rm) const {
        return score_ > rm.score_;
    }
    bool operator==(const RootMove& rm) const {
        return pv_[0] == rm.pv_[0];
    }
    RootMove(const Move move) {
        pv_[0] = move;
        score_ = -ScoreInfinite;
        score_for_learn_ = -ScoreInfinite;
    }
};
