
#pragma once

#include <cstdlib>

#include "bitboard.h"
#include "move.h"

enum Bound : uint32_t {
    BoundNone,
    BoundUpper,
    BoundLower,
    BoundExact = BoundUpper | BoundLower,
};

//u\1Gg
class TTEntry {
    //move15bitł depth8bitł generation5bit 8+6+2ŕ]lۑł
    //data_.b64(0) score:16bit, bound:2bit, depth:14bit, move:23bit, tsumero:1bit, ply1:1bit, pawndrop_mate:1bit, generation:6bit
    //data_.b64(1) key:64bit Ֆʂ̃L[ɎԂ̎𑫂
    Bitboard data_;

public:
    Score score() const {
        return static_cast<Score>(static_cast<int16_t>(static_cast<uint16_t>(ExtBits<0, 16>(data_.b64(0)))));
    }
    Bound bound() const {
        return static_cast<Bound>(ExtBits<16, 2>(data_.b64(0)));
    }
    Depth depth() const {
        return static_cast<Depth>(ExtBits<18, 14>(data_.b64(0)));
    }
    Move move() const {
        return static_cast<Move>(static_cast<int32_t>(ExtBits<32, 23>(data_.b64(0))));
    }
    bool tsumero() const {
        return (data_.b64(0) & 1ULL << 55) != 0;
    }
    bool ply1() const {
        return (data_.b64(0) & 1ULL << 56) != 0;
    }
    bool pawndrop_mate() const {
        return (data_.b64(0) & 1ULL << 57) != 0;
    }
    int generation() const {
        return static_cast<int>(ExtBits<58, 6>(data_.b64(0)));
    }
    uint64_t key() const {
        return data_.b64(1);
    }

    void clear() {
        data_ = ZeroBB;
    }
    TTEntry(__m128i data) : data_(data) {}
    TTEntry() {}
};

class TranspositionTable {
    static constexpr int ClusterSize = 4;

    struct alignas(64) TTCluster {
        TTEntry entries[ClusterSize];
    };

    const uint64_t cluster_num_, mask_;
    TTCluster *const clusters_;
    /*static thread_local */uint64_t generation_;

public:
    void clear() {
        std::memset(clusters_, 0, cluster_num_ * sizeof *clusters_);
    }
    TTEntry *firstEntry(const uint64_t key) const {
        return clusters_[key & mask_].entries;
    }
    void prefetch(const uint64_t key) const {
        _mm_prefetch(reinterpret_cast<char *>(firstEntry(key)), _MM_HINT_T0);
    }
    //݂̋ǖʂ̒u\̃GgԂ(trueԂ)
    bool probe(TTEntry& ret, uint64_t key) const {
        TTEntry *const tte = firstEntry(key);
        for (int i = 0; i < ClusterSize; i++) {
            TTEntry entry = tte[i];
            if (entry.key() == key) {
                ret = entry;
                return true;
            }
        }
        return false;
    }
    //XgAȂƂ move2323bitȂ̂ŕt32bit炢ȂLXgĂv
    void store(uint64_t key64, Score score16, Bound bound2, int depth14, Move move23, uint32_t tsumero1 = 0, uint32_t ply1 = 0, uint32_t pawndrop_mate1 = 0) {
        const TTEntry entry = _mm_setr_epi64x(
            static_cast<uint16_t>(score16) | bound2 << 16 | depth14 << 18
            | static_cast<uint64_t>(move23.value_) << 32 | static_cast<uint64_t>(tsumero1) << 55 | static_cast<uint64_t>(ply1) << 56 | static_cast<uint64_t>(pawndrop_mate1) << 57 | static_cast<uint64_t>(generation_) << 58,
            key64);
        store(entry);
    }
    void store(const TTEntry& entry) {
        const uint64_t key = entry.key();
        TTEntry *const tte = firstEntry(key);
        int v_min = INT_MAX, i_min;
        //keyu\ɂȂƂǂu邩
        for (int i = 0; i < ClusterSize; i++){
            if (tte[i].generation() != generation_ || tte[i].key() == 0) {
                tte[i] = entry;
                return;
            }
            if (tte[i].key() == key) {
                /*if (entry.depth() >= tte[i].depth() || abs(entry.score()) >= ScoreMateInMaxPly) */tte[i] = entry;
                return;
            }
            int v = tte[i].depth() + (tte[i].bound() == BoundExact) * (int)OnePly * 8;
            if (v < v_min){
                v_min = v;
                i_min = i;
            }
        }
        tte[i_min] = entry;
    }
    void incrementGeneration() {
        generation_ = (generation_ + 1) & MaskBits(0, 6);
    }
    int generation() {
        return static_cast<int>(generation_);
    }

    TranspositionTable() : cluster_num_{}, mask_{}, clusters_{} {}

    TranspositionTable(size_t size) :
        cluster_num_{ static_cast<uint32_t>(ToPow2(size / sizeof TTCluster)) },
        mask_{ cluster_num_ - 1 },
        clusters_{ static_cast<TTCluster *>(std::calloc(static_cast<size_t>(cluster_num_), sizeof TTCluster)) },
        generation_{ 0 } {}

    ~TranspositionTable() {
        free(clusters_);
    }
};

extern TranspositionTable tt;
