
#include <iostream>
#include <sstream>
#include <fstream>
#include <new>
#include <set>
#include <vector>
#include <random>
#include <regex>
#include <thread>
#include <filesystem>

#include "usi.h"
#include "thread.h"
#include "search.h"
#include "movegen.h"
#include "tt.h"
#include "evaluate.h"
#include "learn.h"
#include "util.h"

namespace sys = std::tr2::sys;

std::string Mate;
bool Ponder, SmartPvOutput = true;
int Hash = 256, Margin[2], BookPly, game_ply;

int t_min, t_max;

namespace {
    constexpr int MarginBottom = 8, MarginTop = 2, StackSize = MarginBottom + MaxGamePly + MarginTop;
    Position *pos_array = nullptr;
};

//SFEN̋ǖʂposɃZbg
Position *setPosition(Position *ppos, std::istringstream &iss) {
    //startpos̏߂ǂ
    iss.str(std::regex_replace(
        iss.str().substr(static_cast<size_t>(iss.tellg()) + 1),
        std::regex("startpos"), "sfen lnsgkgsnl/1r5b1/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL b - 1",
        std::regex_constants::format_first_only));

    std::string sfenPos, sfenTurn, sfenHand, sfenCount, sfenMove;
    Position& pos = *ppos;

    //ǖʏ
    std::memset(&pos - MarginBottom, 0, MaxGamePly * sizeof pos);
    pos.clear();
    game_ply = 0;

    //Ֆ
    iss >> sfenPos;//"sfen"
    iss >> sfenPos;
    int x = 0, y = 0, pro = 0;
    for (auto c : sfenPos) {
        if ('0' < c && c <= '9') {
            x += c - '0';
        } else if (c == '+') {
            pro = 1;
        } else if (c == '/') {
            x = 0;
            y++;
        } else {
            auto i = sfenPiece.find(c);
            assert(i != std::string::npos && i < PieceTypeNum);

            Color c = (i < HandTypeNum) ? Black : White;
            PieceType p = static_cast<int>(i) % HandTypeNum + pro * HandTypeNum;
            Square sq = square(x, y);

            pos.byColorBB_[c] ^= SquareBB[sq];
            pos.piece_[sq] = p;

            pro = 0;
            x++;
        }
    }

    //
    iss >> sfenTurn;
    pos.root_turn_ = (sfenTurn == "b") ? Black : White;

    //
    iss >> sfenHand;
    int n = 0;
    for (auto c : sfenHand) {
        if (c == '-') {
            break;
        } else if ('0' <= c && c <= '9') {
            n = n * 10 + (c - '0');
        } else {
            auto i = sfenPiece.find(c);
            assert(i != std::string::npos && i < PieceTypeNum);
            pos.hand_[i / HandTypeNum][static_cast<int>(i) % HandTypeNum] = std::max(n, 1);
            n = 0;
        }
    }

    //萔(gȂ)
    iss >> sfenCount;

    pos.normalize();

#ifdef SEE_TEST
    {
        Color c = pos.root_turn_;
        Move moves[MaxMove];
        Move *last = (c == Black) ? generate_pseudo_legal<Black>(moves, pos) : generate_pseudo_legal<White>(moves, pos);
        std::cout << "info string " << pos.toUSI() << std::endl;
        std::for_each(moves, last, [&](Move move) {
            if (move.to() != square(4, 2)) return;
            std::cout << "info string " << move.toUSI() << " : ";
            const int threshold = 250;
            auto s = (c == Black) ? pos.see<Black>(move, threshold) : pos.see<White>(move, threshold);
            std::cout << ": " << s << std::endl;
        });
        std::quick_exit(0);
    }
#endif

    //w
    iss >> sfenMove;
    if (sfenMove != "moves") return &pos;
    
    Color c = pos.root_turn_;
    for (int i = 0;; ++i) {
        Position& pos = ppos->forward(i);
        if (!(iss >> sfenMove)) {
            pos.root_turn_ = c;
            pos.ply_ = 0;
            return &pos;
        }

        //SĂ̍@𐶐Ĉv̂T(񍇖@𐶐ȂƂ͌ĂȂ)
        Move moves[MaxMove];
        Move *last = generate_pseudo_legal(moves, c, pos);

        auto it = std::find_if(moves, last, [&](const Move move) {
            return sfenMove == move.toUSI();
        });

        if (!(it < last)) {
            std::for_each(moves, last, [&](Move move) {
                std::cout << move.toUSI() << " ";
            });
            std::cout << '\n' << last - moves << ' ' << i << ' ' << c << ' ' << pos.toUSI() << std::endl;
        }
        assert(it < last);

        ++game_ply;
        (c == Black) ? pos.moveGivesCheck<Black>(*it) : pos.moveGivesCheck<White>(*it);
        pos.doMove(c, *it);
        (c == Black) ? evaluate<Black>(pos.next()) : evaluate<White>(pos.next());
        c = OppositeColor(c);
    }
}

void init() {
    if (pos_array) return;

    auto read = [](auto& vector, const std::string path) {
        vector.clear();
        std::ifstream ifs{ path, std::ios::in | std::ios::binary };
        if (ifs.is_open()) {
            vector.resize(ifs.seekg(0, std::ios::end).tellg() / sizeof vector[0]);
            ifs.seekg(0, std::ios::beg);
            ifs.read(reinterpret_cast<char *>(&vector[0]), vector.size() * sizeof vector[0]);
            ifs.close();
        } else {
            //ƁAĂÃ^C~OƏɕ\Ȃ
            std::cout << "info string " << path << " not found." << std::endl;
        }
    };

    in_ponder = false;
    init_bitboards();
    init_search_table();
    init_piece_score();
    Position::initZobrist();
    new(&tt) TranspositionTable{ static_cast<size_t>(Hash) * 1024 * 1024 };
    read(v_book, "book.bin");

    read(v_pp, "pp.bin");
    if (v_pp.empty()) v_pp.assign(PPSize + ATSize, 0);
    if (v_pp.size() != PPSize + ATSize) v_pp.resize(PPSize + ATSize, 0);
    p_pp = reinterpret_cast<decltype(p_pp)>(&v_pp[0]);
    //std::memset(p_pp, 0, PPSize * 2);
    p_at = &v_pp[PPSize];

    //for (PIndex pi = 0; pi < PEndIndex; ++pi) {
    //    int sign = (pi < BlackPEndIndex) ? 1 : -1;
    //    PieceType p = IndexToPiece(pi);
    //    int16_t s = (PieceScore[p] * FVScale + 20) / 39 * sign;
    //    for (PIndex i = 0; i < PEndIndex; ++i) {
    //        if (i == pi) continue;
    //        p_pp[pi][i] += s;
    //        p_pp[i][pi] += s;
    //    }
    //}

    //std::cout << "fv_is_ok:" << fv_is_ok() << std::endl;

    pos_array = static_cast<Position *>(std::malloc(StackSize * sizeof *pos_array));
}

//GUĨR}h󂯎
void usiLoop() {
    Position *ppos = nullptr;

    std::string line, token;

    while (std::getline(std::cin, line)) {
        std::istringstream iss(line);
        iss >> token;

        if (token == "usi") {
            std::cout << "id name shogi686_wcsc26";
#ifdef _M_IX86
            std::cout << "_32bit";
#endif
#ifdef DEBUG
            std::cout << "_debug";
#endif
            std::cout << std::endl;
            std::cout << "id author merom686" << std::endl;
            std::cout << "option name Ponder type check default false" << std::endl;
            std::cout << "option name Hash type spin default 1024 min 0 max 32768" << std::endl;
            std::cout << "option name Threads type spin default 0 min 0 max 32" << std::endl;
            std::cout << "option name BookNum type spin default 30 min 0 max 256" << std::endl;
            std::cout << "option name MarginNormal type spin default 100 min 0 max 3000" << std::endl;
            std::cout << "option name MarginByoyomi type spin default 100 min 0 max 15000" << std::endl;
            std::cout << "option name Mate type combo default Mate var Mate var Perft var Reverse" << std::endl;
            std::cout << "option name SmartPvOutput type check default true" << std::endl;
            std::cout << "usiok" << std::endl;

        } else if (token == "setoption") {
            iss >> token;
            if (token == "name") {
                iss >> token;
                if (token == "Mate"){
                    iss >> token;
                    if (token == "value") {
                        iss >> token;
                        Mate = token;
                    }
                } else if (token == "Ponder") {//ԂȂRsy
                    iss >> token;
                    if (token == "value") {
                        iss >> token;
                        Ponder = (token == "true");
                    }
                } else if (token == "Hash") {
                    iss >> token;
                    if (token == "value") {
                        iss >> token;
                        Hash = std::stoi(token);
                    }
                } else if (token == "Threads") {
                    iss >> token;
                    if (token == "value") {
                        iss >> token;
                        thread_num = std::stoi(token);
                        if (thread_num == 0) {
                            thread_num = physical_concurrency();
                        }
                    }
                } else if (token == "BookNum") {
                    iss >> token;
                    if (token == "value") {
                        iss >> token;
                        BookPly = std::stoi(token);
                    }
                } else if (token == "MarginNormal") {
                    iss >> token;
                    if (token == "value") {
                        iss >> token;
                        Margin[0] = std::stoi(token);
                    }
                } else if (token == "MarginByoyomi") {
                    iss >> token;
                    if (token == "value") {
                        iss >> token;
                        Margin[1] = std::stoi(token);
                    }
                } else if (token == "SmartPvOutput") {
                    iss >> token;
                    if (token == "value") {
                        iss >> token;
                        SmartPvOutput = (token == "true");
                    }
                }
            }

        } else if (token == "isready") {
            init();
            std::cout << "readyok" << std::endl;

        } else if (token == "position") {
            assert(in_ponder || !::stop);
            while (in_ponder) { ::stop = true; std::this_thread::sleep_for(std::chrono::milliseconds(1)); }
            _mm_mfence();

            ppos = setPosition(pos_array + MarginBottom, iss);

        } else if (token == "go") {
            iss >> token;
            assert(ppos);

            if (token == "btime") {
                int btime, wtime, byoyomi;
                iss >> btime >> token >> wtime >> token >> byoyomi;
                const bool is_byoyomi = (token == "byoyomi");
                //std::cout << "byo " << is_byoyomi << std::endl;
                int t_remain = (ppos->root_turn_ == Black) ? btime : wtime;//̎c莞

                //̈Ō̂Ƃl[gő̎(ʏ1bȏ)
                t_max = t_remain + byoyomi - Margin[1];
                //̈ōŒႱꂾ͍l(w肪m肵ꍇ͏)
                if (is_byoyomi && t_remain < byoyomi) {
                    t_min = t_max;
                } else {
                    t_min = std::max(t_remain / 64, byoyomi / 2);
                    t_min = std::max((t_min + 999) / 1000 * 1000, 1000) - Margin[0];//؂グ
                    t_min = std::min(t_min, t_max);
                }
                std::cout << "btime:" << btime << " wtime:" << wtime << " byoyomi:" << byoyomi
                    << " t_remain:" << t_remain << " t_min:" << t_min << " t_max:" << t_max << std::endl;
                think(*ppos, t_min);

            } else if (token == "infinite") {
                think(*ppos, 0);

            } else if (token == "mate") {
                if (Mate == "Mate") {
                    std::cout << "checkmate notimplemented" << std::endl;
                } else if (Mate == "Perft"){
                    perft(*ppos);
                }
            }

        } else if (token == "stop") {
            ::stop = true;

        } else if (token == "quit") {
            ::stop = true;
            break;

        } else {
            //(usinewgame, ponderhit, gameover)͕
        }
    }
}

void test(int argc, char *argv[]) {
    init();
    //std::string s = "position sfen R8/2K1S1SSk/4B4/9/9/9/9/9/1L1L1L3 b RBGSNLP3g3n17p 1\n";//593
    //std::string s = "position sfen ln1g3nl/1ks1g1r2/pppp2bp1/4p1p1p/5p1P1/2s1P3P/PP1P1PP2/LSGB3R1/KNS2G1NL w p 1\n";
    //std::string s = "position startpos moves 2g2f 8c8d 2f2e 8d8e 6i7h 4a3b 2e2d 2c2d 2h2d P*2c 2d2h 9c9d 9g9f 3c3d 3i3h 7a6b 3h2g 8e8f 8g8f 8b8f P*8g 8f8d 2g3f 2b3c 1g1f 3a2b 1f1e 5a4a 7g7f 6a5a 4g4f 7c7d 3f4e 8a7c 6g6f 5c5d 4e3d 3c4d 7i6h 7d7e 6h6g 5d5e 3d2e 4d5c 7f7e 5c7e 5i4h 7e5c 8i7g 2a3c 2e3f 8d7d 7g6e 7c6e 6f6e 5e5f 5g5f P*3e 3f4g 5c7e 4h3h 7e5g+ P*7f 7d2d P*2e 3c2e P*2f 2e3g+ 3h3g 2d3d 4i4h 3e3f 4g3f N*2d P*3e 2d3f 3g3f S*4g 4h4g 3d3e 3f3e 5g4g 6g5h G*3f 3e2e 4g4f P*3g 1c1d 1e1d P*1h 3g3f 4f2h N*3d 1h1i+ R*2a L*3a 3d2b+ R*2d 2e3e 3b2b N*3c 2b3c";
    //std::string s = "position startpos moves 2g2f 3c3d 7g7f 8b4b 2f2e 5a6b 5i6h 2b8h+ 7i8h 6b7b 6h7h 7b8b 3i4h 9a9b 4i5h 3a2b 2e2d 2c2d 2h2d B*3h 2d2h 3h4i+ B*1f 4i1f 1g1f 4c4d 8h7g 4a3b 1f1e 4d4e 7h8h 8b9a 5h6h 7a8b 6h7h 6a7a B*5a 4b5b 5a2d+ 3b3c 2d2e 5b4b 2e2f 3c4c P*2d 2a3c 2f1f P*2e 2i1g 4c4d 1g2e 3c2e 2h2e 4d3e 2e2h 4e4f 4g4f 3e4f P*4c 4b6b N*6e B*6d 2d2c+ P*4g 4h3i 4f3g 2h2d 4g4h+ 2c2b 4h3i 2b1a S*7d S*6f 3g4g L*8f N*9d 2d2g 9d8f 8g8f 7d6e 6f6e 6d3g+ 2g3g 4g3g B*4d 6c6d 6e5f 3i3h 4d5c+ R*4i 6i7i L*5b 5c6b 7a6b N*9e 5b5f S*7a 6b7b 7a8b 9a8b S*6a 7b7a R*4b B*6b 5g5f 4i1i+ 6a5b+ S*6a 9e8c+ 8b8c L*8e P*8d 8e8d 8c8d 5b6a 7a6a 1f2f N*9e S*9f 6b2f 4b9b+ 1i7i 8h7i 8d7d 9f9e 2f3e L*6h 7d6c 4c4b+ S*7b N*8d B*8c R*3c L*5c 8d7b+ 8c9b 7b6a R*4i 7i8h 9b5f 3c4c+ 4i4c+ 4b4c 6c5d P*5g 5f4g+ 4c5c 5d4e S*5f 4e3f 5f4g 3g4g B*7b P*4e R*4c R*5e 5g5f 5e9e L*9f S*6i 9f9e G*4d 4c2c 6i7h+ 8h7h G*2d 2c2b+ N*6e S*6f S*5g 6f6e 6d6e 7b8a+ S*5i 5c5d 4d5d 8a5d 5g6h+ 7h8g S*7h 8g9f 6h6g R*4c 4g4f G*5e 7h8i 5e4e N*8d 9f8e 3e1g+ 4e4f 3f3g 4f3f 3g2h 2b2d 1g2f 2d2f L*2g B*1g 2h1i 2f2g P*2h 4c4i+ 3h4i 2g2h";
    std::string s = "position startpos\n";
    std::istringstream iss(s);
    std::string token;
    iss >> token;
    Position *ppos = setPosition(pos_array + MarginBottom, iss);

    std::string cmd = argv[1];
    sys::path dir;
    if (argc > 2) dir.assign(argv[2]);

    auto for_each_file = [](const sys::path dir, const std::string ext, auto f) {
        auto to_lower = [](std::string s) {
            std::transform(s.cbegin(), s.cend(), s.begin(), tolower);
            return s;
        };
        std::for_each(sys::recursive_directory_iterator{ dir }, sys::recursive_directory_iterator{}, [&](const sys::path &path) {
            if (!sys::is_regular_file(path) || to_lower(path.extension().string()) != ext) return;

            std::ifstream ifs{ path };
            if (!ifs.is_open()) {
                std::cout << "can't open.\n" << path << std::endl;
                return;
            }

            std::string str(std::istreambuf_iterator<char>{ifs}, std::istreambuf_iterator<char>{});
            ifs.close();

            f(path, str);
        });
    };


    if (cmd == "search") {
        std::cout << sizeof Position << " " << sizeof PositionMin << std::endl;

        const int msec = 2000;
        think(*ppos, msec);
        std::this_thread::sleep_for(std::chrono::milliseconds(msec + 100));
        std::quick_exit(0);

    } else if (cmd == "perft") {
        perft(*ppos);
        std::this_thread::sleep_for(std::chrono::hours(168));

    //} else if (cmd == "reverse") {
    //    reverse(*ppos);
    //    std::this_thread::sleep_for(std::chrono::hours(168));

    } else if (cmd == "floodgate") {
        //܂floodgateCSAt@CA+[eBOɂ Blunder.ConverterSFENɕϊ(蓮)
        dir.append("floodgate");
        int n = 0, m = 0;
        for_each_file(dir, ".csa", [&](const sys::path path, const std::string str) {
            // 2
            int winner = 3;
            //҂̃[eBO
            int rate[2];

            for (Color c = Black; c < ColorNum; ++c) {
                static const std::regex re{ std::string{ (c == Black) ? "black" : "white" } + R"(_rate:\S+:(\S+))" };
                std::smatch m;
                rate[c] = std::regex_search(str, m, re) ? std::max(std::stoi(m[1]), 0) : 1;
            }
            auto p1 = str.rfind("win");
            auto p2 = str.rfind("lose");
            if (p1 != std::string::npos && p2 != std::string::npos) {
                if (str.rfind("toryo:") != std::string::npos || str.rfind("kachi:") != std::string::npos) winner = (p1 > p2);
            } else {
                auto p3 = str.rfind("draw:");
                if (p3 != std::string::npos) winner = 2;
            }
            if (m % 1000 == 0) std::cout << n << " / " << m << std::endl;
            ++m;
            if (winner >= 2) return;
            if (rate[winner] < 3000 && rate[winner ^ 1] < 3000) return;

            std::string new_name = std::to_string(10000 + n++) + "_" + std::to_string(winner) + "_"
                + std::to_string(rate[0]) + "_" + std::to_string(rate[1]) + ".csa";
            sys::path new_path = dir.parent_path().append("floodgate_CSA").append(new_name);
            sys::rename(path, new_path);
            //std::cout << new_path << std::endl;
        });
        std::cout << n << " / " << m << std::endl;

    } else if (cmd == "book" || cmd == "learn") {
        //book:SFENǂݍŒՂ
        //learn:SFENǂݍŋ@BwK
        dir.append("sfen");
        const bool cmd_is_book = (cmd == "book");

        for (int k = 0; k < 2; ++k) {
            std::set<BookEntry> book_entry_set;
            std::vector<PositionLearn> v_posl;
            int n = 0;
            for_each_file(dir, ".sfen", [&](const sys::path path, const std::string str) {
                static const std::regex re{ R"((\d+)_(\d)_(\d+)_(\d+))" };
                std::smatch m;
                std::string filename = path.filename().string();
                if (!std::regex_search(filename, m, re)) {
                    std::cout << "invalid filename.\n" << path << std::endl;
                    return;
                }
                if (++n % 1000 == 0) std::cout << n << std::endl;

                int data[3];
                for (int i = 0; i < 3; ++i) data[i] = std::stoi(m[i + 2]);
                int winner = data[0];
                int rate[2] = { data[1], data[2] };
                if (cmd_is_book) {
                    if (rate[winner] < 3000 && rate[winner ^ 1] < 3000) return;
                } else {
                    if (rate[winner] < 3000 && rate[winner ^ 1] < 3000) return;
                }

                std::istringstream iss("position " + str + "\n");
                std::string token;
                iss >> token;
                Position *ppos = setPosition(pos_array + MarginBottom, iss);
                Color c = ppos->root_turn_;//2O̎

                for (; !(--ppos)->previous_move_.isNone(); c = OppositeColor(c)) {
                    //if (c != winner) continue;
                    if (cmd_is_book && ppos->ply_ > 50) continue;//Ղ50܂

                    BookEntry book_entry;
                    book_entry.key = (ppos - 1)->key(c);
                    if (k == 0) {
                        book_entry.move = ppos->previous_move_.noScore();
                        book_entry.count = 1;//setɂȂꍇ̂܂insert邽
                        auto it = book_entry_set.find(book_entry);
                        if (it == book_entry_set.end()) {
                            book_entry_set.insert(book_entry);
                        } else {
                            it->count++;
                        }

                    } else {
                        //nbVǖʂ𕜌łȂ̂2
                        auto range = std::equal_range(v_book.begin(), v_book.end(), book_entry, BookEntry::lessKey);
                        if (range.first == range.second || range.first->move.isNone()) continue;

                        PositionLearn posl;
                        std::memcpy(posl.byColorBB_, (ppos - 1)->byColorBB_, sizeof posl.byColorBB_);
                        std::memcpy(posl.hand64_, (ppos - 1)->hand64_, sizeof posl.hand64_);
                        std::memcpy(posl.piece_, (ppos - 1)->piece_, sizeof posl.piece_);
                        posl.root_turn_ = c;
                        std::memset(posl.learn_move_, 0, sizeof posl.learn_move_);

                        int i = 0;
                        std::for_each(range.first, range.second, [&](const BookEntry& a1) {
                            assert(!a1.move.isNone());
                            if (i < _countof(posl.learn_move_)) {
                                posl.learn_move_[i++] = a1.move;
                            }
                        });
                        range.first->move = Move::None();//ǖʂx͓o^Ȃ
                        v_posl.push_back(posl);
                    }
                }
            });

            if (k == 0) {
                v_book.clear();
                std::copy(book_entry_set.begin(), book_entry_set.end(), std::back_inserter(v_book));
                std::sort(v_book.begin(), v_book.end(), BookEntry::lessKeyGreaterCount);//ǖʂ̒ŏo񐔂Ɏwבւ
                std::cout << v_book.size() << std::endl;
            }
            if (cmd_is_book) {
                std::ofstream ofs("book.bin", std::ios::out | std::ios::binary | std::ios::trunc);
                if (!ofs.is_open()) return;
                ofs.write(reinterpret_cast<char *>(&v_book[0]), v_book.size() * sizeof v_book[0]);
                ofs.close();

            } else {
                if (k == 0) continue;
                std::cout << sizeof v_posl[0] << " " << v_posl.size() << std::endl;
                std::free(pos_array);
                learn(std::move(v_posl));
            }
            break;
        }

    } else {
        std::cout << "unknown command : " << cmd << std::endl;
    }
}
